#ifndef _FIREFLY_SHA256_H_
#define _FIREFLY_SHA256_H_

#include "crypto/crypto_types.h"

#include "stl/stdint.hpp"
#include "stl/stddef.hpp"

typedef struct sha256_ctx {
    stl::UInt32T _hash[8];
    stl::UInt8T _data[64];
    stl::SizeT _data_len;
    stl::SizeT _bit_len;
} sha256_ctx;

struct sha256_hmac_ctx {
    struct sha256_ctx _sha_ctx;
    stl::UInt8T _key[SIZEOF(struct sha256_ctx, _data)];
    stl::SizeT _key_length;
    stl::UInt8T _digest[SIZEOF(struct sha256_ctx, _hash)];
};

void sha256_init(struct sha256_ctx* const ctx);
void sha256_update(struct sha256_ctx* const ctx, const stl::UInt8T* input, stl::SizeT length);
const stl::UInt8T* sha256_finish(struct sha256_ctx* const ctx);

void sha256_hmac_init(struct sha256_hmac_ctx* const ctx, const stl::UInt8T* key
    , const stl::SizeT key_length);
const stl::UInt8T* sha256_hmac_finish(struct sha256_hmac_ctx* const ctx
    , const char* message, const stl::SizeT message_length);

void sha224_init(struct sha256_ctx* const ctx);
void sha224_update(struct sha256_ctx* const ctx, const stl::UInt8T* input, stl::SizeT length);
const stl::UInt8T* sha224_finish(struct sha256_ctx* const ctx);

#endif
