#ifndef _FIREFLY_SHA512_H_
#define _FIREFLY_SHA512_H_

#include "crypto/crypto_types.h"

#include "stl/stdint.hpp"
#include "stl/stddef.hpp"

struct sha512_ctx {
    stl::UInt64T _hash[8];
    stl::UInt8T _data[128];
    size64_t _data_len;
    size128_t _bit_len;
};

struct sha512_hmac_ctx {
    struct sha512_ctx _sha_ctx;
    stl::UInt8T _key[SIZEOF(struct sha512_ctx, _hash)];
    size64_t _key_length;
    stl::UInt8T _digest[SIZEOF(struct sha512_ctx, _hash)];
};

void sha512_init(struct sha512_ctx* const ctx);
void sha512_update(struct sha512_ctx* const ctx, const stl::UInt8T* input, size64_t length);
const stl::UInt8T* sha512_finish(struct sha512_ctx* const ctx);

void sha512_hmac_init(struct sha512_hmac_ctx* const ctx, const stl::UInt8T* key
    , const size128_t key_length);
const stl::UInt8T* sha512_hmac_finish(struct sha512_hmac_ctx* const ctx
    , const char* message, const size128_t message_length);

#endif
