#ifndef _LOGGING_HPP_
#define _LOGGING_HPP_

#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if (defined STL_ENABLE_LOGGING && (STL_ENABLE_LOGGING == 1))

#include "stl/compiler.hpp"
#include "stl/ostream_types.hpp"
#include "stl/ostream.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"
#include "stl/time.hpp"

namespace logging {
enum class LvlE : stl::SizeT
{
  Info,
  Warning,
  Error,
  Success,
  Exception,
  Debug,
  _Count
};

enum class StatE : stl::SizeT
{
  Success,
  Fail,
  Warning,
  _Count
};

// TODO: Log to file

class Logger : public stl::stream::OStream
{
public:
  Logger() = default;

  virtual ~Logger() {}

  template<typename T>
  Logger& operator<<(T* const value)
  {
    OStream::operator<<(value);
    return *this;
  }

  template<typename T>
  Logger& operator<<(const T& value)
  {
    OStream::operator<<(value);
    return *this;
  }

  virtual stl::SizeT write(const stl::UInt8T* const data, stl::SizeT size) override;
};

extern Logger log; // Default logger
}

#define LOGEXT                                                                 \
  (logging::log << "[" << __FILENAME__ << " #" << __LINE__ << ": "             \
                << stl::time::DateTime() << "] ")

#endif // (defined STL_ENABLE_LOGGING && (STL_ENABLE_LOGGING == 1))

#endif
