#ifndef __STL_NETWORK_PACKET_BUILDER_HPP__
#define __STL_NETWORK_PACKET_BUILDER_HPP__

#include "stl/endian.hpp"
#include "stl/memory.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"
#include "stl/vector.hpp"

namespace stl::network {
using PacketDataT = Vector<ByteT>;

template<EndianE Endianess>
class Packet;

template<EndianE Endianess>
class PacketManip
{
public:
  PacketManip() = default;
  virtual ~PacketManip() = default;

  virtual void update(Packet<Endianess>& packet) const = 0;

private:
};

template<EndianE Endianess>
class Packet
{
public:
  Packet() = default;

  Packet(const Packet& other) { *this = other; }

  Packet& operator=(const Packet& other) { m_data = other.m_data; }

  Packet(Packet&& other)
    : m_data(::stl::move(other.m_data))
  {
  }

  SizeT size() const { return m_data.size(); }

  const ByteT* data() const { return m_data.data(); }

  template<typename T>
  void insert(const T& value, const SizeT offset)
  {
    Assert(offset <= m_data.size());

    const T temp = Endian::hostSync(Endianess, value);
    const ByteT* ptr = reinterpret_cast<const ByteT*>(&temp);

    for (SizeT i = 0U; i < sizeof(T); ++i) {
      m_data.insert(m_data.begin() + offset + i, *(ptr + i));
    }
  }

  template<typename T>
  Packet& operator<<=(const T& value)
  {
    const T temp = Endian::hostSync(Endianess, value);
    const ByteT* ptr = reinterpret_cast<const ByteT*>(&temp);

    for (SizeT i = 0U; i < sizeof(T); ++i) {
      m_data.push_back(*(ptr + i));
    }

    return *this;
  }

  Packet& operator<<=(const stl::UInt8T& value)
  {
    m_data.push_back(value);

    return *this;
  }

  Packet& operator<<=(const stl::Int8T& value)
  {
    m_data.push_back(static_cast<stl::UInt8T>(value));

    return *this;
  }

  Packet& operator<<=(const String& value)
  {
    const ByteT* ptr = reinterpret_cast<const ByteT*>(value.c_str());

    for (SizeT i = 0U; i < value.length(); ++i) {
      m_data.push_back(*(ptr + i));
    }

    return *this;
  }

  Packet& operator<<=(const Vector<ByteT>& value)
  {
    const ByteT* ptr = reinterpret_cast<const ByteT*>(value.data());

    for (SizeT i = 0U; i < value.size(); ++i) {
      m_data.push_back(*(ptr + i));
    }

    return *this;
  }

  Packet& operator<<=(const char* value)
  {
    Assert(value);

    const ByteT* ptr = reinterpret_cast<const ByteT*>(value);

    while (0U != *ptr) {
      m_data.push_back(*(ptr++));
    }

    return *this;
  }

  operator PacketDataT() {
    return ::stl::move(m_data);
  }

private:
  PacketDataT m_data;
};
}

#endif // __STL_NETWORK_PACKET_BUILDER_HPP__
