#ifndef __STL_PROTOCOLS_MODBUS_HPP__
#define __STL_PROTOCOLS_MODBUS_HPP__

#include "stl/stddef.hpp"
#include "stl/stdint.hpp"

namespace stl::protocol::modbus {
class ModbusStream
{
public:
private:
  // PLACINg THE LRC INTO THE MESSAgE
  // When the the 8–bit LRC (2 ASCII characters) is transmitted in the message,
  // the high–order character will be transmitted first, followed by the
  // low–order character. For example, if the LRC value is $52 (0101 0010):
  // Colon
  // ‘:’
  // Addr Func Data
  // Count
  // Data Data …. Data LRC
  // Hi ‘5’
  // LRC
  // Lo‘2’
  // CR LF

  // *frame – pointer on “Addr” of message (pucFrame)
  // usLen – length message from “Addr” to end “Data”
  static stl::UInt8T calcLRC(stl::UInt8T* frame, stl::SizeT length)
  {
    stl::UInt8T lrc = 0U; /* LRC char initialized */

    while (length--) {
      lrc += *frame++; /* Add buffer byte without carry */
    }

    /* Return twos complement */
    lrc = static_cast<stl::UInt8T>(-static_cast<stl::Int8T>(lrc));

    return lrc;
  }

  static stl::UInt16T calcCRC16(stl::UInt8T* data, stl::SizeT length)
  {
    /* ModBus_CRC16 Calculatd CRC16 with polynome 0xA001 and init value 0xFFFF
    Input *data - pointer on data
    Input Lenght - number byte in buffer
    Output - calculated CRC16
    */
    stl::UInt16T output = 0xFFFFU;

    do {
      stl::UInt16T i = 8;

      output = output ^ *data++;

      do {
        if (0x0001U & output) {
          output >>= 1;
          output ^= 0xA001U;
        } else {
          output >>= 1;
        }
      } while (--i);
    } while (--length);

    return output;
  }
};
}

#endif // __STL_PROTOCOLS_MODBUS_HPP__
