#ifndef __STL_ALLOCATE_HPP__
#define __STL_ALLOCATE_HPP__

#include "stl/assert.hpp"
#include "stl/stddef.hpp"
#include "stl/utility.hpp"

namespace stl::memory
{
   class MemoryManagerBase
   {
   public:
      static SizeT allocated();

   protected:
      static SizeT m_allocated;

      // static void* allocateArrayInternal(SizeT size);
      // static void deallocateArrayInternal(void* ptr);

    private:
      MemoryManagerBase();
      ~MemoryManagerBase();
   };

   template<typename T>
   class MemoryManager : public MemoryManagerBase
   {
   public:
      static T* allocate()
      {
         // T* ptr = static_cast<T*>(allocateArrayInternal(sizeof(T)));
         // return new (ptr) T();
         return new T();
      }

      template<typename ... ArgsT>
      static T* allocate(ArgsT&&... args)
      {
         // void* ptr = static_cast<T*>(allocateArrayInternal(sizeof(T)));

         // return new (reinterpret_cast<T*>(ptr)) T(forward<ArgsT>(args)...);

         return new T(forward<ArgsT>(args)...);
      }

      static void deallocate(T* const ptr)
      {
         // deallocateArrayInternal(ptr);
         delete ptr;
      }

   private:
      MemoryManager() = default;
      ~MemoryManager();
   };

   template<typename T>
   class MemoryManager<T[]> : public MemoryManagerBase
   {
   private:
      using BareT = RemoveExtentT<T>;

      MemoryManager() = default;
      ~MemoryManager();

   public:
      static BareT* allocate(SizeT count)
      {
         /* BareT* ptr = static_cast<BareT*>(allocateArrayInternal(sizeof(BareT) * count));

         for(SizeT i = 0U; i < count; ++i)
         {
            new (&ptr[i]) BareT();
         }

         return reinterpret_cast<BareT*>(ptr); */

         return new BareT[count]();
      }

      static void deallocate(BareT* const ptr)
      {
         // deallocateArrayInternal(ptr); // TODO: Implement correctly for platforms where allocation is not thread safe

         delete [] ptr;
      }

   private:

   };
} // namespace stl::memory

#endif // __STL_ALLOCATE_HPP__
