#ifndef _STL_ARRAY_HPP_
#define _STL_ARRAY_HPP_

#include "stl/assert.hpp"
#include "stl/stddef.hpp"

namespace stl {
  template<typename T, SizeT Size>
  class Array {
  public:
    Array() = default;

    Array(const Array& other) {
      for(SizeT index = 0U; index < Size; ++index) {
        m_data[index] = other.m_data[index];
      }
    }

    Array(Array&& other) = delete;

    Array& operator = (const Array& other) {
      for(SizeT index = 0U; index < Size; ++index) {
        m_data[index] = other.m_data[index];
      }

      return *this;
    }

    Array& operator = (Array&& other) = delete;

    static constexpr SizeT maxSize() {
      return Size;
    }

    SizeT size() const {
      return Size;
    }

    bool empty() const {
      return Size == 0U;
    }

    T* data() {
      return m_data;
    }

    const T* data() const {
      return m_data;
    }

    T& front() {
      Assert(Size > 0U);
      return m_data[0];
    }

    const T& front() const {
      Assert(Size > 0U);
      return m_data[0];
    }

    T& back() {
      Assert(Size > 0U);
      return m_data[Size - 1];
    }

    const T& back() const {
      Assert(Size > 0U);
      return m_data[Size - 1];
    }

    T& operator [](const SizeT index) {
      Assert((Size > 0U) && (index < Size));
      return m_data[index];
    }

    const T& operator [](const SizeT index) const {
      Assert((Size > 0U) && (index < Size));
      return m_data[index];
    }

    void fill(const T& value) {
      for(SizeT i = 0U; i < Size; ++i) {
        m_data[i] = value;
      }
    }

  private:
    T m_data[Size];
  };
}

#endif // _STL_ARRAY_HPP_
