#ifndef __STL_ATOMIC_HPP__
#define __STL_ATOMIC_HPP__

#include "stl/mutex.hpp"
#include "stl/stddef.hpp"
#include "stl/stl_config.hpp"
#include "stl/utility.hpp"

namespace stl {
template<typename T>
class Atomic
{
public:
  Atomic() : m_lock("atomic") {}

  Atomic(const T& value)
    : m_value(value) {}

  Atomic(T&& value)
    : m_value(::stl::move(value)) {}

  Atomic(const Atomic& other) = delete;

  Atomic& operator=(const Atomic& other) = delete;

  Atomic(Atomic&& other) = delete;

  Atomic& operator=(Atomic&& other) = delete;

  T load() const
  {
    T temp;

    {
      LockGuard<Mutex> lock(m_lock);
      temp = m_value;
    }

    return temp;
  }

  void store(T&& value)
  {
    LockGuard<Mutex> lock(m_lock);
    m_value = move(value);
  }

  void store(const T& value)
  {
    LockGuard<Mutex> lock(m_lock);
    m_value = value;
  }

  operator T() const { return load(); }

private:
  mutable Mutex m_lock;
  T m_value = static_cast<T>(0);
};
}

#endif // __STL_ATOMIC_HPP__
