#ifndef _COMPILE_INFO_HPP_
#define _COMPILE_INFO_HPP_

// ############################################################################
// # OS                                                                       #
// ############################################################################

// Windows
// ************************************

#if defined _WIN64 || defined _WIN32
#ifndef _WINSOCKAPI_
#define _WINSOCKAPI_
#include <Windows.h>
#undef _WINSOCKAPI_
#else
#include <Windows.h>
#endif
#endif

#if defined _WIN64
#define CI_OS_WINDOWS 1
#define CI_SUPPORT_STD 1
#elif defined _WIN32
#define CI_OS_WINDOWS 1
#define CI_SUPPORT_STD 1
#else
#define CI_OS_WINDOWS 0
#endif

// Unix
// ************************************

#if defined unix || defined __unix || defined __unix__
#define CI_OS_UNIX 1
#define CI_SUPPORT_STD 1
#else
#define CI_OS_UNIX 0
#endif

// Linux
// ************************************

#if defined __linux__ ||                                                       \
  defined __gnu_linux__ // linux && __linux not POSIX compilant
#define CI_OS_LINUX 1
#define CI_SUPPORT_STD 1
#else
#define CI_OS_LINUX 0
#endif

// Android
// ************************************

#if defined __ANDROID__
#define CI_OS_ANDROID 1
#define CI_SUPPORT_STD 1
#else
#define CI_OS_ANDROID 0
#endif

// No OS
// ************************************

#if (defined CI_OS_WINDOWS && CI_OS_WINDOWS == 1) ||                           \
  (defined CI_OS_UNIX && CI_OS_UNIX == 1) ||                                   \
  (defined CI_OS_LINUX && CI_OS_LINUX == 1) ||                                 \
  (defined CI_OS_ANDROID && CI_OS_ANDROID == 1)
#define CI_OS_NONE 0
#else
#define CI_OS_NONE 1
#endif

// ############################################################################
// # CUSTOM PLATFORM                                                          #
// ############################################################################

#if defined(CI_CUSTOM_PLATFORM)
#if  CI_CUSTOM_PLATFORM == 9
#define CI_PLATFORM_MKRWIFI1010 1
#define CI_PLATFORM_ARDUINO 1
#define CI_REAL_GPIO 1
#elif CI_CUSTOM_PLATFORM == 10
#define CI_PLATFORM_PC_TEST 1
#define CI_PLATFORM_PC 1
#define CI_REAL_GPIO 1
#elif CI_CUSTOM_PLATFORM == 11
#define CI_PLATFORM_RPI_ZEROW 1
#define CI_PLATFORM_RPI 1
#define CI_REAL_GPIO 1
#elif CI_CUSTOM_PLATFORM == 12
#define CI_PLATFORM_RPI_ZERO2W 1
#define CI_PLATFORM_RPI 1
#define CI_REAL_GPIO 1
#elif CI_CUSTOM_PLATFORM == 15
#define CI_PLATFORM_RPI_ZERO3B 1
#define CI_PLATFORM_RPI 1
#define CI_REAL_GPIO 1
#endif
#endif

// ############################################################################
// # PLATFORM                                                                 #
// ############################################################################

#if defined(__x86_64__) || defined(_M_X64)
#define CI_PLATFORM_X86_64 1
#elif defined(i386) || defined(__i386__) || defined(__i386) || defined(_M_IX86)
#define CI_PLATFORM_X86_32 1
#elif defined(__ARM_ARCH_2__)
#define CI_PLATFORM_ARM2 1
#elif defined(__ARM_ARCH_3__) || defined(__ARM_ARCH_3M__)
#define CI_PLATFORM_ARM3 1
#elif defined(__ARM_ARCH_4T__) || defined(__TARGET_ARM_4T)
#define CI_PLATFORM_ARM4T 1
#elif defined(__ARM_ARCH_5_) || defined(__ARM_ARCH_5E_)
#define CI_PLATFORM_ARM5 1
#elif defined(__ARM_ARCH_6T2_) || defined(__ARM_ARCH_6T2_)
#define CI_PLATFORM_ARM6T2 1
#elif defined(__ARM_ARCH_6__) || defined(__ARM_ARCH_6J__) ||                   \
  defined(__ARM_ARCH_6K__) || defined(__ARM_ARCH_6Z__) ||                      \
  defined(__ARM_ARCH_6ZK__)
#define CI_PLATFORM_ARM6 1
#elif defined(__ARM_ARCH_7__) || defined(__ARM_ARCH_7A__) ||                   \
  defined(__ARM_ARCH_7R__) || defined(__ARM_ARCH_7M__) ||                      \
  defined(__ARM_ARCH_7S__)
#define CI_PLATFORM_ARM7 1
#elif defined(__ARM_ARCH_7A__) || defined(__ARM_ARCH_7R__) ||                  \
  defined(__ARM_ARCH_7M__) || defined(__ARM_ARCH_7S__)
#define CI_PLATFORM_ARM7A 1
#elif defined(__ARM_ARCH_7R__) || defined(__ARM_ARCH_7M__) ||                  \
  defined(__ARM_ARCH_7S__)
#define CI_PLATFORM_ARM7R 1
#elif defined(__ARM_ARCH_7M__)
#define CI_PLATFORM_ARM7M 1
#elif defined(__ARM_ARCH_7S__)
#define CI_PLATFORM_ARM7S 1
#elif defined(__aarch64__) || defined(_M_ARM64)
#define CI_PLATFORM_ARM64 1
#elif defined(mips) || defined(__mips__) || defined(__mips)
#define CI_PLATFORM_MIPS 1
#elif defined(__sh__)
#define CI_PLATFORM_SUPERH 1
#elif defined(__powerpc) || defined(__powerpc__) || defined(__powerpc64__) ||  \
  defined(__POWERPC__) || defined(__ppc__) || defined(__PPC__) ||              \
  defined(_ARCH_PPC)
#define CI_PLATFORM_POWERPC_32 1
#elif defined(__PPC64__) || defined(__ppc64__) || defined(_ARCH_PPC64)
#define CI_PLATFORM_POWERPC_64 1
#elif defined(__sparc__) || defined(__sparc)
#define CI_PLATFORM_SPARC 1
#elif defined(__m68k__)
#define CI_PLATFORM_M68K 1
#else
#define CI_PLATFORM_UNKNOWN 1
#endif

// Arduino
// ************************************

// ESP32
// ************************************

// ESP8266
// ************************************

// Defaults

#if !defined(CI_REAL_GPIO)
#define CI_REAL_GPIO 0
#endif

#if !defined(CI_PLATFORM_RPI)
#define CI_PLATFORM_RPI 0
#endif

#if !defined(CI_PLATFORM_PC)
#define CI_PLATFORM_PC 0
#endif

#if !defined(CI_PLATFORM_ARDUINO)
#define CI_PLATFORM_ARDUINO 0
#endif

// ############################################################################
// # PLATFORM ENDIAN                                                          #
// ############################################################################

#define CI_PLATFORM_ENDIAN_BIG 2
#define CI_PLATFORM_ENDIAN_LITTLE 1
#define CI_PLATFORM_ENDIAN_UNSUPPORTED 0

#if (defined(__BYTE_ORDER) && (__BYTE_ORDER == __BIG_ENDIAN)) ||               \
  defined(__BIG_ENDIAN__) || defined(__ARMEB__) || defined(__THUMBEB__) ||     \
  defined(__AARCH64EB__) || defined(_MIBSEB) || defined(__MIBSEB) ||           \
  defined(__MIBSEB__)
#define CI_PLATFORM_ENDIAN CI_PLATFORM_ENDIAN_BIG
#elif (defined(__BYTE_ORDER) && (__BYTE_ORDER == __LITTLE_ENDIAN)) ||          \
  defined(__LITTLE_ENDIAN__) || defined(__ARMEL__) || defined(__THUMBEL__) ||  \
  defined(__AARCH64EL__) || defined(_MIPSEL) || defined(__MIPSEL) ||           \
  defined(__MIPSEL__)
#define CI_PLATFORM_ENDIAN CI_PLATFORM_ENDIAN_LITTLE
#else
#if defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1) && defined REG_DWORD
#if REG_DWORD_LITTLE_ENDIAN == REG_DWORD
#define CI_PLATFORM_ENDIAN CI_PLATFORM_ENDIAN_LITTLE
#elif REG_DWORD_BIG_ENDIAN == REG_DWORD
#define CI_PLATFORM_ENDIAN CI_PLATFORM_ENDIAN_BIG
#else
#define CI_PLATFORM_ENDIAN CI_PLATFORM_ENDIAN_UNSUPPORTED
#endif
#else
#define CI_PLATFORM_ENDIAN CI_PLATFORM_ENDIAN_UNSUPPORTED
#endif
#endif

// ############################################################################
// # COMPILER                                                                 #
// ############################################################################

#if defined(__clang__)
#define CI_COMPILER_CLANG 1
#elif defined(__GNUC__) || defined(__GNUG__)
#define CI_COMPILER_GNUC 1
#elif defined(_MSC_VER)
#define CI_COMPILER_MSC 1
#else
#error Unsupported compiler
#endif

#if !defined CI_SUPPORT_STD
#define CI_SUPPORT_STD 0
#endif // !defined CI_SUPPORT_STD

// ############################################################################
// # OTHER                                                                    #
// ############################################################################

// TODO

#endif
