#ifndef _STL_COMPILER_HPP_
#define _STL_COMPILER_HPP_

#include "stl/compile_info.hpp"

#define CFLT(value) static_cast<float>(value)
#define CDBL(value) static_cast<double>(value)
#define CU08(value) static_cast<stl::UInt8T>(value)
#define CI08(value) static_cast<stl::Int8T>(value)
#define CU16(value) static_cast<stl::UInt16T>(value)
#define CI16(value) static_cast<stl::Int16T>(value)
#define CU32(value) static_cast<stl::UInt32T>(value)
#define CI32(value) static_cast<stl::Int32T>(value)
#define CU64(value) static_cast<stl::UInt64T>(value)
#define CI64(value) static_cast<stl::Int64T>(value)

#ifndef UNUSED_PARAM
#define UNUSED_PARAM(param)                                                    \
  {                                                                            \
    (void)(param);                                                             \
    (void)&(param);                                                            \
  }
#endif

#if defined(CI_COMPILER_MSC) && (CI_COMPILER_MSC == 1)
#include <string.h>

#define __FILENAME__                                                           \
  (strrchr(__FILE__, '\\') ? strrchr(__FILE__, '\\') + 1     \
                                    : __FILE__)
#else
#define __FILENAME__                                                           \
  (__builtin_strrchr(__FILE__, '/') ? __builtin_strrchr(__FILE__, '/') + 1     \
                                    : __FILE__)
#endif

#define CALL_REPORT() stl::cout << __FILENAME__ << ':' << __LINE__ << stl::endl;
#define NOP

#if defined CI_COMPILER_GNUC && (CI_COMPILER_GNUC == 1)
#define C_PACK(__Declaration__) __Declaration__ __attribute__((__packed__))
#endif

#if defined CI_COMPILER_MSC && (CI_COMPILER_MSC == 1)
#define C_PACK(__Declaration__)                                                \
  __pragma(pack(push, 1)) __Declaration__ __pragma(pack(pop))
#endif

struct Impl
{
  void* _impl;
};

namespace stl {
template<typename TO, typename TI>
TO
cast(const TI value)
{
  return static_cast<TO>(value);
}
} // namespace stl

#endif
