#ifndef _LIMITS_HPP_
#define _LIMITS_HPP_

#define NOMINMAX

#if defined min
#undef min
#endif

#if defined max
#undef max
#endif

#include "stl/stl_config.hpp"

#include <limits.h>

#define NOMINMAX

#define IMAX_BITS(m) ((m) /((m)%0x3fffffffL+1) /0x3fffffffL %0x3fffffffL *30 \
                  + (m)%0x3fffffffL /((m)%31+1)/31%31*5 + 4-12/((m)%31+3))

// TODO: Se according to compiler macros  __UINT32_MAX__ ...

namespace stl {
    template<typename T>
    class Limits;

    template<>
    class Limits<signed char> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr signed char min() noexcept {
            return -128;
        }

        [[nodiscard]] static constexpr signed char max() noexcept {
            return 127;
        }
    };

    template<>
    class Limits<char> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr char min() noexcept {
            return -128;
        }

        [[nodiscard]] static constexpr char max() noexcept {
            return 127;
        }
    };

    template<>
    class Limits<unsigned char> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr unsigned char min() noexcept {
            return 0U;
        }

        [[nodiscard]] static constexpr unsigned char max() noexcept {
            return 255U;
        }
    };

    template<>
    class Limits<short> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr short min() noexcept {
            return -32768;
        }

        [[nodiscard]] static constexpr short max() noexcept {
            return 32767;
        }
    };

    template<>
    class Limits<unsigned short> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr unsigned short min() noexcept {
            return 0U;
        }

        [[nodiscard]] static constexpr unsigned short max() noexcept {
            return 65535U;
        }
    };

    template<>
    class Limits<int> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr int min() noexcept {
            return sizeof(int) == 2 ? -32768 : -2147483648L;
        }

        [[nodiscard]] static constexpr int max() noexcept {
            return sizeof(int) == 2 ? 32767 : 2147483647L;
        }
    };

    template<>
    class Limits<unsigned int> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr unsigned int min() noexcept {
            return 0UL;
        }

        [[nodiscard]] static constexpr unsigned int max() noexcept {
            return sizeof(unsigned int) == 2 ? 65535U : 4294967295UL;
        }
    };

    template<>
    class Limits<long> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr long min() noexcept {
            return sizeof(long) == 2U ? -32768L : -2147483648L;
        }

        [[nodiscard]] static constexpr long max() noexcept {
            return sizeof(long) == 2U ? 32767L : -2147483647L;
        }
    };

    template<>
    class Limits<unsigned long> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr unsigned long min() noexcept {
            return sizeof(unsigned long) == 2U ? 0 : 0UL;
        }

        [[nodiscard]] static constexpr unsigned long max() noexcept {
            return sizeof(unsigned long) == 2U ? 65535UL : 4294967295UL;
        }
    };

#if defined STL_ENABLE_INT64
    static_assert(sizeof(unsigned long long int) == 8, "Unexpected UInt64 size");

    template<>
    class Limits<long long int> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr long long int min() noexcept {
            return -9223372036854775807LL;
        }

        [[nodiscard]] static constexpr long long int max() noexcept {
            return 9223372036854775807LL;
        }
    };

    template<>
    class Limits<unsigned long long int> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr unsigned long long int min() noexcept {
            return 0ULL;
        }

        [[nodiscard]] static constexpr unsigned long long int max() noexcept {
            return 18446744073709551615ULL;
        }
    };
#endif

#if defined STL_ENABLE_FLOAT
    template<>
    class Limits<float> {
    public:
        Limits() = delete;

#if defined STL_ENABLE_FLOAT
        [[nodiscard]] static constexpr float min() noexcept {
            return 1.175494e-38f;
        }

        [[nodiscard]] static constexpr float max() noexcept {
            return 3.402823e+38f;
        }

        [[nodiscard]] static constexpr float epsilon() noexcept {
            return 1.401298E-45f;
        }
#endif
    };
#endif

#if defined STL_ENABLE_DOUBLE
    template<>
    class Limits<double> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr double min() noexcept {
            return 2.225074e-308;
        }

        [[nodiscard]] static constexpr double max() noexcept {
            return 1.797693e+308;
        }

        [[nodiscard]] static constexpr double epsilon() noexcept {
            return 4.94065645841247E-324;
        }
    };
#endif

#if defined STL_ENABLE_LONG_DOUBLE
    template<>
    class Limits<long double> {
    public:
        Limits() = delete;

        [[nodiscard]] static constexpr long double min() noexcept {
#error "min value not defined yet"
        }

        [[nodiscard]] static constexpr long double max() noexcept {
#error "max value not defined yet"
        }

        [[nodiscard]] static constexpr long double epsilon() noexcept {
#error "epsilon value not defined yet"
        }
    };
#endif
}

#endif
