#ifndef __STL_MUTEX_HPP__
#define __STL_MUTEX_HPP__

#include "stl/functional.hpp"
#include "stl/memory.hpp"
#include "stl/string.hpp"

namespace stl {
class Lockable
{
public:
  virtual void lock() = 0;
  virtual bool tryLock() = 0;
  virtual void unlock() = 0;
};

class Mutex : public Lockable
{
public:
  Mutex();
  Mutex(String name);

  virtual ~Mutex();

  Mutex(const Mutex&) = delete;
  Mutex& operator=(const Mutex&) = delete;

  Mutex(Mutex&& other) = delete;
  Mutex& operator=(Mutex&& other) = delete;

  virtual void lock() override;
  virtual void unlock() override;
  virtual bool tryLock() override;

private:
  class MutexImpl;

  UniquePtr<MutexImpl> m_handleContainer;
  String m_name;
};

template<typename T>
class LockGuard final
{
public:
  LockGuard(Lockable& lockable)
    : m_ref(lockable)
  {
    m_ref.get().lock();
  }

  LockGuard(const LockGuard&) = delete;
  LockGuard& operator=(const LockGuard&) = delete;

  LockGuard(LockGuard&& other) = delete;
  LockGuard& operator=(LockGuard&& other) = delete;

  ~LockGuard() { m_ref.get().unlock(); }

private:
  ReferenceWrapper<Lockable> m_ref;
};

} // namespace stl

#endif // __STL_MUTEX_HPP__
