#ifndef _STRING_HPP_
#define _STRING_HPP_

#include "stl/assert.hpp"
#include "stl/limits.hpp"
#include "stl/memory.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/stl_config.hpp"
#include "stl/string_utils.hpp"

#include <string>

// TODO: Optimize string (reserved == size right now) + UTF8 string impl

namespace stl {
using String2 = std::string;

/* class String2 : public std::string
{
public:
  String2() = default;
  virtual ~String2() = default;

  String2(const String2& other)
    : std::string(static_cast<const std::string&>(other))
  {}

  String2(String2&& other)
    : std::string(stl::move(static_cast<std::string&>(other)))
  {}

  String2(const std::string& other)
    : std::string(other)
  {}

  String2(std::string&& other)
    : std::string(stl::move(other))
  {}

  String2(const char* str, const stl::SizeT size)
    : std::string(str, size)
  {}

  

  String2& operator = (const std::string& other)
  {
    std::string::operator = (other);

    return *this;
  }

  String2& operator = (std::string&& other)
  {
    std::string::operator = (stl::move(other));

    return *this;
  }

  String2& operator = (const String2& other)
  {
    std::string::operator = (static_cast<const std::string&>(other));

    return *this;
  }

  String2& operator = (String2&& other)
  {
    std::string::operator = (stl::move(static_cast<std::string&>(other)));

    return *this;
  }

private:
}; */

class String
{
public:
  static constexpr SizeT nPos = Limits<SizeT>::max();

  String();

  String(const String& other);

  String(String&& other) noexcept;

  String(const char* const str);

  String(const char* const str, const SizeT length);

  String(const char* const str, const SizeT offset, const SizeT length);

  String(const char c, const SizeT count);

  virtual ~String() = default;

  // Getters

  const SizeT reserved() const noexcept;

  const SizeT size() const noexcept;

  const SizeT length() const noexcept;

  bool empty() const noexcept;

  const char* c_str() const;

  const char* data() const { return m_ptr.get(); }

  // Allocation methods

  void resize(const SizeT newSize);

  void reserve(const SizeT toReserve);

  // Growth methods

  // Main implementation of string growth
  void insert(const SizeT lOffset,
              const char* what,
              const SizeT rOffset,
              const SizeT length);

  void insert(const char* what, const SizeT rOffset, const SizeT length);

  void insert(const char* what, const SizeT length);

  void insert(const char* what);

  void insert(const SizeT lOffset,
              const String& what,
              const SizeT rOffset,
              const SizeT length);

  void insert(const String& what, const SizeT rOffset, const SizeT length);

  void insert(const String& what, const SizeT length);

  void insert(const String& what);

  void append(const char* what, const SizeT rOffset, const SizeT length);

  void append(const char* what, const SizeT length);

  void append(const char* what);

  void append(const char what);

  void append(const String& what, const SizeT rOffset, const SizeT length);

  void append(const String& what, const SizeT length);

  void append(const String& what);

  void append(String&& what);

  // Shrink methods

  void remove(const SizeT offset, const SizeT length);

  bool clear();

  // Operators

  String& operator=(const String& other);

  String& operator=(String&& other);

  String& operator=(const char* const str);

  String operator+(const char c) const;

  String& operator+=(const char c);

  String operator+(const char* str) const;

  String& operator+=(const char* str);

  String operator+(const String& string) const;

  String& operator+=(const String& string);

  String operator+(String&& string) const;

  String& operator+=(String&& string);

  char& operator[](const SizeT index);

  const char& operator[](const SizeT index) const;

  // Utility methods

  int csCompare(const char* str) const;

  int csCompare(const SizeT offset, const char* str) const;

  int csCompare(const SizeT offset, const char* str, SizeT toCompare) const;

  int ciCompare(const char* str) const;

  int ciCompare(const SizeT offset, const char* str) const;

  int ciCompare(const SizeT offset, const char* str, SizeT toCompare) const;

  int csCompare(const String& str) const;
  int csCompare(const SizeT offset, const String& str) const;
  int csCompare(const SizeT offset,
                const String& str,
                const SizeT toCompare) const;

  int ciCompare(const String& str) const;
  int ciCompare(const SizeT offset, const String& str) const;
  int ciCompare(const SizeT offset,
                const String& str,
                const SizeT toCompare) const;

  String substr(const SizeT offset, const SizeT length) const;

  String substr(
    const SizeT offset) const; // Add const char* start, const char* end

  SizeT find(const char c) const;

  SizeT find(const char* str) const;

  SizeT csFind(const char* str, const SizeT length) const;

  SizeT ciFind(const char* str, const SizeT length) const;

  SizeT csFind(const char* str) const;

  SizeT ciFind(const char* str) const;

  SizeT ciFind(const String& str, const SizeT length) const;

  SizeT ciFind(const String& str) const;

  SizeT replace(const char what, const char with);

  bool operator==(const String& other) const;

  bool operator!=(const String& other) const;

  bool operator>(const String& other) const;

  bool operator>=(const String& other) const;

  bool operator<(const String& other) const;

  bool operator<=(const String& other) const;

private:
  UniquePtr<char[]> m_ptr;
  SizeT m_size;
  SizeT m_reserved;
};

inline void upper(String2& str)
{
  for(SizeT i = 0U; i < str.length(); ++i)
  {
    str[i] = string::toUpper(str[i]);
  }
}

inline void lower(String2& str)
{
  for(SizeT i = 0U; i < str.length(); ++i)
  {
    str[i] = string::toLower(str[i]);
  }
}
} // namespace stl

#endif
