#ifndef __STL_STRING_CONV_HPP__
#define __STL_STRING_CONV_HPP__

#include "stl/stddef.hpp"
#include "stl/stl_config.hpp"
#include "stl/string.hpp"

#include "log/logging.hpp"

namespace stl::string {
extern char g_intConvTable[72];

template<typename T>
String
toString(const T& value, const T base)
{
  T temp = value;

  String output;

  bool negative = false;

  if (temp < 0) {
    negative = true;
    temp *= static_cast<T>(-1);
  }

  // check that the base if valid
  if (base < static_cast<T>(2) || base > static_cast<T>(36)) {
    return "";
  }

  T tmp_value;

  do {
    tmp_value = temp;
    temp /= base;
    output += g_intConvTable[35 + (tmp_value - temp * base)];
  } while (temp);

  if (output.length() > 0U) {
    if (negative) {
      output += '-';
    }

    reverse(&output[0], &output[output.length() - 1]);
  }

  return output;
}

template<typename T>
String
toString(const T& value)
{
  return toString(value, static_cast<T>(10));
}

String
toString(const float& value);
String
toString(const double& value);

template<typename T>
bool
toString(const T& value, const T base, char* output, SizeT outputMaxSize)
{
  T temp = value;

  SizeT numberLength = 0U;

  // check that the base if valid
  if (base < static_cast<T>(2) || base > static_cast<T>(36)) {
    return false;
  }

  bool negative = false;

  if (temp < 0) {
    negative = true;
    temp *= static_cast<T>(-1);
  }

  T tmp_value;

  do {
    if (numberLength >= outputMaxSize) {
      return false;
    }

    tmp_value = temp;
    temp /= base;
    *(output + (numberLength++)) =
      g_intConvTable[35 + (tmp_value - temp * base)];
  } while (temp);

  if (numberLength > 0U) {
    if (negative) {
      if ((numberLength + 1) >= outputMaxSize) {
        return false;
      }

      *(output + (numberLength++)) = '-';
    }

    reverse(&output[0], &output[numberLength - 1]);
  }

  if (numberLength >= outputMaxSize) {
    return false;
  }

  output[numberLength] = '\0';

  return true;
}

template<typename T>
bool
toString(const T& value, char* output, SizeT outputMaxSize)
{
  return toString(value, static_cast<T>(10), output, outputMaxSize);
}

template<typename T>
T
stringTo(const char* string, SizeT length, bool& result, SizeT& numberLength)
{
  numberLength = 0U;

  if (0U == length) {
    result = true;
    return static_cast<T>(0);
  }

  T output = static_cast<T>(0);
  const char* ptr = string;

  result = false;

  bool negative = false;
  if ('-' == *ptr) {
    negative = true;
    ++ptr;
    --length;
    ++numberLength;
  }

  if (length == 0U) {
    result = false;
    return static_cast<T>(0);
  }

  while (*ptr && length--) {
    if (!isDigit(*ptr)) {
      break;
    }

    output = output * static_cast<T>(10U) + (*ptr++ - '0');
    ++numberLength;
  }

  if (numberLength == (negative ? 1U : 0U)) {
    result = false;
    return static_cast<T>(0);
  }

  if (negative) {
    output *= static_cast<T>(-1);
  }

  result = true;

  return output;
}

template<typename T>
T
stringTo(const String& string, bool& result, SizeT& numberLength)
{
  return stringTo<T>(string.c_str(), string.length(), result, numberLength);
}

template<typename T>
T
stringTo(const SizeT offset,
        const String& string,
         bool& result,
         SizeT& numberLength)
{
  Assert(offset < string.length());
  return stringTo<T>(string.c_str() + offset, string.length() - offset, result, numberLength);
}

template<typename T>
T
stringTo(const String& string, SizeT length, bool& result, SizeT& numberLength)
{
  return stringTo<T>(string.c_str(),
                     length > string.length() ? string.length() : length,
                     result,
                     numberLength);
}

template<typename T>
T
stringTo(const String& string,
         const SizeT offset,
         const SizeT length,
         bool& result,
         SizeT& numberLength)
{
  Assert(length <= string.length());
  Assert(offset < string.length());
  Assert((offset + length) <= string.length());

  return stringTo<T>(string.c_str() + offset, length, result, numberLength);
}

String
toHex(const UInt8T* data, const SizeT size, bool upperCase);

template<typename T>
T fromHex(String hex)  // TODO: Refactor, return error 
{ 
    if(hex.length() == 0U)
    {
        return 0;
    } 
  
    // Initializing base value to 1, i.e 16^0 
    T base = 1; 
  
    T dec = 0; 
  
    // Extracting characters as digits from last 
    // character 
    for (stl::SizeT i = hex.length(); i > 0; --i) {
        if (hex[i - 1] >= '0' && hex[i - 1] <= '9') { 
            dec += (int(hex[i - 1]) - 48) * base; 
  
            // incrementing base by power 
            base = base * 16; 
        }
        else if (hex[i - 1] >= 'A' && hex[i - 1] <= 'F') { 
            dec += (int(hex[i - 1]) - 55) * base; 
  
            // incrementing base by power 
            base = base * 16; 
        } 
        else if (hex[i - 1] >= 'a' && hex[i - 1] <= 'f') { 
            dec += (int(hex[i - 1]) - 55) * base; 
  
            // incrementing base by power 
            base = base * 16; 
        } 
    } 
    return dec; 
}
} // namespace stl::string
#endif
