#ifndef _STRING_UTILS_HPP_
#define _STRING_UTILS_HPP_

#include "stl/assert.hpp"
#include "stl/limits.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/stl_config.hpp"

namespace stl
{
class String;

namespace string
{
inline bool isAlpha(const stl::UInt8T b)
{
    return ((static_cast<char>(b) >= 'a') && (static_cast<char>(b) <= 'z')) ||
           ((static_cast<char>(b) >= 'A') && (static_cast<char>(b) <= 'Z'));
}

inline bool isDigit(const stl::UInt8T b)
{
    return ((static_cast<char>(b) >= '0') && (static_cast<char>(b) <= '9'));
}

inline bool isXDigit(const stl::UInt8T b)
{
    return isDigit(b) ||
           ((static_cast<char>(b) >= 'a') && (static_cast<char>(b) <= 'f')) ||
           ((static_cast<char>(b) >= 'A') && (static_cast<char>(b) <= 'F'));
}

inline bool isAlnum(const stl::UInt8T b) { return isAlpha(b) || isDigit(b); }

inline bool isPrint(const stl::UInt8T b)
{
    return ((static_cast<char>(b) >= ' ') && (static_cast<char>(b) <= '~'));
}

inline bool isCtrl(const stl::UInt8T b)
{
    return ((static_cast<char>(b) >= '\0') && (static_cast<char>(b) <= 0x1F));
}

inline bool isSpace(const stl::UInt8T b)
{
    return ((static_cast<char>(b) >= 0x09) && (static_cast<char>(b) <= 0x0D)) ||
           (static_cast<char>(b) == ' ');
}

inline bool isBlank(const stl::UInt8T b)
{
    return (static_cast<char>(b) == '\t') || (static_cast<char>(b) == ' ');
}

inline bool isUpper(const char c) { return (c >= 'a') && (c >= 'z'); }

inline bool isLower(const char c) { return (c >= 'A') && (c >= 'Z'); }

inline char toUpper(const char c)
{
    Assert(c >= '\0');
    return ((c >= 'a') && (c <= 'z')) ? (c - 0x20) : c;
}

inline char toLower(const char c)
{
    Assert(c >= '\0');
    return ((c >= 'A') && (c <= 'Z')) ? (c + 0x20) : c;
}

inline SizeT length(const char* str)
{
    const char* const begin = str;
    while ('\0' != *str)
    {
        ++str;
    }

    return static_cast<SizeT>(str - begin);
}

char* copy(const char* begin, const char* const end, char* destination,
           const SizeT destinationLength);

char nibbleToHex(UInt8T nibble, const bool upperCase);

void reverse(char* strFront, char* strBack);

void reverse(char* str, SizeT length);

void reverse(char* str);

bool endsWith(const String& input, const char* with);

bool endsWith(const String& input, const String& with);

bool startsWith(const String& input, const char* with);

bool startsWith(const String& input, const String& with);
} // namespace string
} // namespace stl

#endif
