#ifndef _TIME_HPP_
#define _TIME_HPP_

#include "stl/location.hpp"
#include "stl/math.hpp"
#include "stl/stdint.hpp"
#include "stl/stl_config.hpp"

#include "log/logging.hpp"

#if defined STL_ENABLE_INT64
namespace stl {
namespace time {
// using TimeT = stl::UInt64T;
using STimeT = stl::Int64T;

class Millis;

Millis
getUpTime();
Millis
getUTCTimestamp();

class Nanos;
class Micros;
class Millis;
class Sec;
class Mins;
class Hours;

class TimeValue
{
public:
  TimeValue();

  TimeValue(const TimeValue& other);

  TimeValue& operator=(const TimeValue& other);

  TimeValue(TimeValue&& other);

  TimeValue& operator=(TimeValue&& other);

  TimeValue& operator=(const Nanos& value);
  TimeValue& operator=(const Micros& value);
  TimeValue& operator=(const Millis& value);
  TimeValue& operator=(const Sec& value);
  TimeValue& operator=(const Mins& value);
  TimeValue& operator=(const Hours& value);

  TimeValue(STimeT value);

  TimeValue operator+(const TimeValue& other) const;

  TimeValue operator+=(const TimeValue& other)
  {
    m_timeValue += other.m_timeValue; // TODO Check for overflow?

    return *this;
  }

  TimeValue operator-(const TimeValue& other) const;

  TimeValue operator-=(const TimeValue& other)
  {
    m_timeValue -= other.m_timeValue; // TODO Check for underflow?

    return *this;
  }

  bool operator==(const TimeValue& other) const;

  bool operator!=(const TimeValue& other) const;

  bool operator>(const TimeValue& other) const;

  bool operator<(const TimeValue& other) const;

  bool operator>=(const TimeValue& other) const;

  bool operator<=(const TimeValue& other) const;

  STimeT baseValue() const { return m_timeValue; }

protected:
  STimeT m_timeValue;
};

class Nanos : public TimeValue
{
public:
  static constexpr STimeT MULTIPLIER = static_cast<STimeT>(1LL);

  Nanos() = default;

  Nanos(STimeT value);

  Nanos(const TimeValue& timeValue);

  Nanos(const Nanos& other)
    : TimeValue(static_cast<const TimeValue&>(other))
  {
  }

  Nanos& operator=(const Nanos& other)
  {
    static_cast<TimeValue&>(*this) = static_cast<const TimeValue&>(other);

    return *this;
  }

  Nanos operator+(const Nanos& other) const
  {
    return TimeValue(m_timeValue + other.m_timeValue);
  }

  Nanos operator-(const Nanos& other) const
  {
    return TimeValue(m_timeValue - other.m_timeValue);
  }

  bool operator==(const STimeT value) const { return this->value() == value; }

  bool operator!=(const STimeT value) const { return this->value() != value; }

  bool operator>(const STimeT value) const { return this->value() > value; }

  bool operator<(const STimeT value) const { return this->value() < value; }

  bool operator>=(const STimeT value) const { return this->value() >= value; }

  bool operator<=(const STimeT value) const { return this->value() <= value; }

  bool operator==(const Nanos& other) const
  {
    return m_timeValue == other.m_timeValue;
  }

  bool operator!=(const Nanos& other) const
  {
    return m_timeValue != other.m_timeValue;
  }

  bool operator>(const Nanos& other) const
  {
    return m_timeValue > other.m_timeValue;
  }

  bool operator<(const Nanos& other) const
  {
    return m_timeValue < other.m_timeValue;
  }

  bool operator>=(const Nanos& other) const
  {
    return m_timeValue >= other.m_timeValue;
  }

  bool operator<=(const Nanos& other) const
  {
    return m_timeValue <= other.m_timeValue;
  }

  STimeT value() const;
};

class Micros : public TimeValue
{
public:
  static constexpr STimeT MULTIPLIER = static_cast<STimeT>(1000LL);

  Micros() = default;

  Micros(const TimeValue& timeValue);

  Micros(STimeT value);

  Micros(const Micros& other)
    : TimeValue(static_cast<const TimeValue&>(other))
  {
  }

  Micros& operator=(const Micros& other)
  {
    static_cast<TimeValue&>(*this) = static_cast<const TimeValue&>(other);

    return *this;
  }

  Micros operator+(const Micros& other) const
  {
    return TimeValue(m_timeValue + other.m_timeValue);
  }

  Micros operator-(const Micros& other) const
  {
    return TimeValue(m_timeValue - other.m_timeValue);
  }

  bool operator==(const STimeT value) const { return this->value() == value; }

  bool operator!=(const STimeT value) const { return this->value() != value; }

  bool operator>(const STimeT value) const { return this->value() > value; }

  bool operator<(const STimeT value) const { return this->value() < value; }

  bool operator>=(const STimeT value) const { return this->value() >= value; }

  bool operator<=(const STimeT value) const { return this->value() <= value; }

  bool operator==(const Micros& other) const
  {
    return m_timeValue == other.m_timeValue;
  }

  bool operator!=(const Micros& other) const
  {
    return m_timeValue != other.m_timeValue;
  }

  bool operator>(const Micros& other) const
  {
    return m_timeValue > other.m_timeValue;
  }

  bool operator<(const Micros& other) const
  {
    return m_timeValue < other.m_timeValue;
  }

  bool operator>=(const Micros& other) const
  {
    return m_timeValue >= other.m_timeValue;
  }

  bool operator<=(const Micros& other) const
  {
    return m_timeValue <= other.m_timeValue;
  }

  STimeT value() const;
};

class Millis : public TimeValue
{
public:
  static constexpr STimeT MULTIPLIER = static_cast<STimeT>(1000000LL);

  Millis() = default;

  Millis(const TimeValue& timeValue);

  Millis(STimeT value);

  Millis(const Millis& other)
    : TimeValue(static_cast<const TimeValue&>(other))
  {
  }

  Millis& operator=(const Millis& other)
  {
    static_cast<TimeValue&>(*this) = static_cast<const TimeValue&>(other);

    return *this;
  }

  Millis operator+(const Millis& other) const
  {
    return TimeValue(m_timeValue + other.m_timeValue);
  }

  Millis operator-(const Millis& other) const
  {
    return TimeValue(m_timeValue - other.m_timeValue);
  }

  Millis operator%(const STimeT value)
  {
    return TimeValue(m_timeValue % (value * MULTIPLIER));
  }

  Millis& operator%=(const STimeT value)
  {
    m_timeValue %= (value * MULTIPLIER);

    return *this;
  }

  bool operator==(const STimeT value) const { return this->value() == value; }

  bool operator!=(const STimeT value) const { return this->value() != value; }

  bool operator>(const STimeT value) const { return this->value() > value; }

  bool operator<(const STimeT value) const { return this->value() < value; }

  bool operator>=(const STimeT value) const { return this->value() >= value; }

  bool operator<=(const STimeT value) const { return this->value() <= value; }

  bool operator==(const Millis& other) const
  {
    return m_timeValue == other.m_timeValue;
  }

  bool operator!=(const Millis& other) const
  {
    return m_timeValue != other.m_timeValue;
  }

  bool operator>(const Millis& other) const
  {
    return m_timeValue > other.m_timeValue;
  }

  bool operator<(const Millis& other) const
  {
    return m_timeValue < other.m_timeValue;
  }

  bool operator>=(const Millis& other) const
  {
    return m_timeValue >= other.m_timeValue;
  }

  bool operator<=(const Millis& other) const
  {
    return m_timeValue <= other.m_timeValue;
  }

  STimeT value() const;
};

class Sec : public TimeValue
{
public:
  static constexpr STimeT MULTIPLIER = static_cast<STimeT>(1000000000LL);

  Sec() = default;

  Sec(const TimeValue& timeValue);

  Sec(STimeT value);

  Sec(const Sec& other)
    : TimeValue(static_cast<const TimeValue&>(other))
  {
  }

  Sec& operator=(const Sec& other)
  {
    static_cast<TimeValue&>(*this) = static_cast<const TimeValue&>(other);

    return *this;
  }

  Sec operator+(const Sec& other) const
  {
    return TimeValue(m_timeValue + other.m_timeValue);
  }

  Sec operator-(const Sec& other) const
  {
    return TimeValue(m_timeValue - other.m_timeValue);
  }

  bool operator==(const STimeT value) const { return this->value() == value; }

  bool operator!=(const STimeT value) const { return this->value() != value; }

  bool operator>(const STimeT value) const { return this->value() > value; }

  bool operator<(const STimeT value) const { return this->value() < value; }

  bool operator>=(const STimeT value) const { return this->value() >= value; }

  bool operator<=(const STimeT value) const { return this->value() <= value; }

  bool operator==(const Sec& other) const
  {
    return m_timeValue == other.m_timeValue;
  }

  bool operator!=(const Sec& other) const
  {
    return m_timeValue != other.m_timeValue;
  }

  bool operator>(const Sec& other) const
  {
    return m_timeValue > other.m_timeValue;
  }

  bool operator<(const Sec& other) const
  {
    return m_timeValue < other.m_timeValue;
  }

  bool operator>=(const Sec& other) const
  {
    return m_timeValue >= other.m_timeValue;
  }

  bool operator<=(const Sec& other) const
  {
    return m_timeValue <= other.m_timeValue;
  }

  STimeT value() const;
};

class Mins : public TimeValue
{
public:
  static constexpr STimeT MULTIPLIER = static_cast<STimeT>(60000000000LL);

  Mins() = default;

  Mins(const TimeValue& timeValue);

  Mins(STimeT value);

  Mins(const Mins& other)
    : TimeValue(static_cast<const TimeValue&>(other))
  {
  }

  Mins& operator=(const Mins& other)
  {
    static_cast<TimeValue&>(*this) = static_cast<const TimeValue&>(other);

    return *this;
  }

  Mins operator+(const Mins& other) const
  {
    return TimeValue(m_timeValue + other.m_timeValue);
  }

  Mins operator-(const Mins& other) const
  {
    return TimeValue(m_timeValue - other.m_timeValue);
  }

  bool operator==(const STimeT value) const { return this->value() == value; }

  bool operator!=(const STimeT value) const { return this->value() != value; }

  bool operator>(const STimeT value) const { return this->value() > value; }

  bool operator<(const STimeT value) const { return this->value() < value; }

  bool operator>=(const STimeT value) const { return this->value() >= value; }

  bool operator<=(const STimeT value) const { return this->value() <= value; }

  bool operator==(const Mins& other) const
  {
    return m_timeValue == other.m_timeValue;
  }

  bool operator!=(const Mins& other) const
  {
    return m_timeValue != other.m_timeValue;
  }

  bool operator>(const Mins& other) const
  {
    return m_timeValue > other.m_timeValue;
  }

  bool operator<(const Mins& other) const
  {
    return m_timeValue < other.m_timeValue;
  }

  bool operator>=(const Mins& other) const
  {
    return m_timeValue >= other.m_timeValue;
  }

  bool operator<=(const Mins& other) const
  {
    return m_timeValue <= other.m_timeValue;
  }

  STimeT value() const;
};

class Hours : public TimeValue
{
public:
  static constexpr STimeT MULTIPLIER = static_cast<STimeT>(3600000000000LL);

  Hours() = default;

  Hours(const TimeValue& timeValue);

  Hours(STimeT value);

  Hours(const Hours& other)
    : TimeValue(static_cast<const TimeValue&>(other))
  {
  }

  Hours& operator=(const Hours& other)
  {
    static_cast<TimeValue&>(*this) = static_cast<const TimeValue&>(other);

    return *this;
  }

  Hours operator+(const Hours& other) const
  {
    return TimeValue(m_timeValue + other.m_timeValue);
  }

  Hours operator-(const Hours& other) const
  {
    return TimeValue(m_timeValue - other.m_timeValue);
  }

  bool operator==(const STimeT value) const { return this->value() == value; }

  bool operator!=(const STimeT value) const { return this->value() != value; }

  bool operator>(const STimeT value) const { return this->value() > value; }

  bool operator<(const STimeT value) const { return this->value() < value; }

  bool operator>=(const STimeT value) const { return this->value() >= value; }

  bool operator<=(const STimeT value) const { return this->value() <= value; }

  bool operator==(const Hours& other) const
  {
    return m_timeValue == other.m_timeValue;
  }

  bool operator!=(const Hours& other) const
  {
    return m_timeValue != other.m_timeValue;
  }

  bool operator>(const Hours& other) const
  {
    return m_timeValue > other.m_timeValue;
  }

  bool operator<(const Hours& other) const
  {
    return m_timeValue < other.m_timeValue;
  }

  bool operator>=(const Hours& other) const
  {
    return m_timeValue >= other.m_timeValue;
  }

  bool operator<=(const Hours& other) const
  {
    return m_timeValue <= other.m_timeValue;
  }

  STimeT value() const;
};

class DateTime
{
public:
  DateTime(Millis utcTimestamp, const UInt32T timeZone);

  DateTime();

  DateTime(const DateTime& other)
    : m_utcTimestamp(other.m_utcTimestamp)
    , m_timeZone(other.m_timeZone)
    , m_sec(other.m_sec)
    , m_min(other.m_min)
    , m_hour(other.m_hour)
    , m_monthDay(other.m_monthDay)
    , m_month(other.m_month)
    , m_year(other.m_year)
    , m_weekDay(other.m_weekDay)
    , m_yearDay(other.m_yearDay)
    , m_isdst(other.m_isdst)
    , m_valid(other.m_valid)
  {
  }

  DateTime(DateTime&& other) = delete;

  DateTime& operator=(const DateTime& other)
  {
    m_utcTimestamp = other.m_utcTimestamp;
    m_timeZone = other.m_timeZone;
    m_sec = other.m_sec;
    m_min = other.m_min;
    m_hour = other.m_hour;
    m_monthDay = other.m_monthDay;
    m_month = other.m_month;
    m_year = other.m_year;
    m_weekDay = other.m_weekDay;
    m_yearDay = other.m_yearDay;
    m_isdst = other.m_isdst;
    m_valid = other.m_valid;

    return *this;
  }

  DateTime& operator=(DateTime&& other) = delete;

  bool valid() const { return m_valid; }

  STimeT seconds() const { return m_sec; }

  STimeT minutes() const { return m_min; }

  STimeT hours() const { return m_hour; }

  STimeT month() const { return m_month; }

  STimeT year() const { return m_year; }

  STimeT monthDay() const { return m_monthDay; }

  STimeT yearDay() const { return m_yearDay; }

  bool isDayLightSavingTime() const { return 0 != m_isdst; }

  SizeT weekDay() const { return m_weekDay; }

  STimeT utcTimeZoneToMs(const UInt32T timeZone);

private:
  static bool utcTimestampToTime(const STimeT utcTimestamp,
                                 DateTime& destination);

  STimeT m_utcTimestamp;
  UInt32T m_timeZone;

  STimeT m_sec;
  STimeT m_min;
  STimeT m_hour;
  STimeT m_monthDay;
  STimeT m_month;
  STimeT m_year;
  SizeT m_weekDay;
  STimeT m_yearDay;
  STimeT m_isdst;

  bool m_valid;
};

class Sun
{
public:
  static Mins sunrise(const CountryE location, const DateTime& date)
  {
    Assert(date.monthDay() > 0U);
    Assert(date.monthDay() <= 31);
    Assert(date.month() > 0);
    Assert(date.month() <= 12);

    static const stl::time::STimeT sunriseTable[12][31] = {
      { 420, 420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
        420, 420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
        420, 420, 420, 420, 420, 420, 420, 420, 420 },
      { 420, 420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
        420, 420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
        420, 420, 420, 420, 420, 420, 420, 420, 420 },
      { 330, 330, 330, 330, 330, 330, 330, 330, 330, 330, 330,
        330, 330, 330, 330, 330, 330, 330, 330, 330, 330, 330,
        330, 330, 330, 330, 330, 330, 330, 330, 330 },
      { 330, 330, 330, 330, 330, 330, 330, 330, 330, 330, 330,
        330, 330, 330, 330, 330, 330, 330, 330, 330, 330, 330,
        330, 330, 330, 330, 330, 330, 330, 330, 330 },
      { 300, 300, 300, 300, 300, 300, 300, 300, 300, 300, 300,
        300, 300, 300, 300, 300, 300, 300, 300, 300, 300, 300,
        300, 300, 300, 300, 300, 300, 300, 300, 300 },
      { 270, 270, 270, 270, 270, 270, 270, 270, 270, 270, 270,
        270, 270, 270, 270, 270, 270, 270, 270, 270, 270, 270,
        270, 270, 270, 270, 270, 270, 270, 270, 270 },
      { 270, 270, 270, 270, 270, 270, 270, 270, 270, 270, 270,
        270, 270, 270, 270, 270, 270, 270, 270, 270, 270, 270,
        270, 270, 270, 270, 270, 270, 270, 270, 270 },
      { 330, 330, 330, 330, 330, 330, 330, 330, 330, 330, 330,
        330, 330, 330, 330, 330, 330, 330, 330, 330, 330, 330,
        330, 330, 330, 330, 330, 330, 330, 330, 330 },
      { 360, 360, 360, 360, 360, 360, 360, 360, 360, 360, 360,
        360, 360, 360, 360, 360, 360, 360, 360, 360, 360, 360,
        360, 360, 360, 360, 360, 360, 360, 360, 360 },
      { 360, 360, 360, 360, 360, 360, 360, 360, 360, 360, 360,
        360, 360, 360, 360, 360, 360, 360, 360, 360, 360, 360,
        360, 360, 360, 360, 360, 360, 360, 360, 360 },
      { 420, 420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
        420, 420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
        420, 420, 420, 420, 420, 420, 420, 420, 420 },
      { 450, 450, 450, 450, 450, 450, 450, 450, 450, 450, 450,
        450, 450, 450, 450, 450, 450, 450, 450, 450, 450, 450,
        450, 450, 450, 450, 450, 450, 450, 450, 450 }
    };

    return stl::time::Mins(sunriseTable[date.month() - 1][date.monthDay() - 1]);
  }

  static Mins sunset(const CountryE location, const DateTime& date)
  {
    Assert(date.monthDay() > 0U);
    Assert(date.monthDay() <= 31);
    Assert(date.month() > 0);
    Assert(date.month() <= 12);

    static const stl::time::STimeT sunsetTable[12][31] = {
      { 960, 960, 960, 960, 960, 960, 960, 960, 960, 960, 960,
        960, 960, 960, 960, 960, 960, 960, 960, 960, 960, 960,
        960, 960, 960, 960, 960, 960, 960, 960, 960 },
      { 1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020,
        1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020,
        1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020, 1020 },
      { 1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050,
        1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050,
        1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050, 1050 },
      { 1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170,
        1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170,
        1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170, 1170 },
      { 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
        1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
        1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200 },
      { 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230,
        1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230,
        1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230 },
      { 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230,
        1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230,
        1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230, 1230 },
      { 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
        1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
        1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200 },
      { 1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140,
        1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140,
        1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140, 1140 },
      { 1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080,
        1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080,
        1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080, 1080 },
      { 960, 960, 960, 960, 960, 960, 960, 960, 960, 960, 960,
        960, 960, 960, 960, 960, 960, 960, 960, 960, 960, 960,
        960, 960, 960, 960, 960, 960, 960, 960, 960 },
      { 960, 960, 960, 960, 960, 960, 960, 960, 960, 960, 960,
        960, 960, 960, 960, 960, 960, 960, 960, 960, 960, 960,
        960, 960, 960, 960, 960, 960, 960, 960, 960 }
    };

    return stl::time::Mins(sunsetTable[date.month() - 1][date.monthDay() - 1]);
  }

  static bool daytime(const CountryE location, const DateTime& date)
  {
    const Mins sunriseTime = sunrise(location, date);
    const Mins sunsetTime = sunset(location, date);

    const Mins timeSinceMidnight = date.hours() * 60 + date.minutes();

    /* logging::log << logging::LvlE::Debug << "Sunrise: " <<
       sunriseTime.value()
                 << " Sunset: " << sunsetTime.value()
                 << " Time since midnight: " << timeSinceMidnight.value()
                 << stl::endl; */

    return timeSinceMidnight >= sunriseTime && timeSinceMidnight < sunsetTime;
  }
};
} // namespace time
} // namespace stl

#else
#error Time library not supported
#endif

#endif
