#ifndef _TYPE_TRAITS_HPP_
#define _TYPE_TRAITS_HPP_

#include "stl/stl_config.hpp"
#include "stl/stddef.hpp"

namespace stl {
    template<typename T, bool Val>
    struct IntegralConstant {
        static constexpr T m_value = Val;

        constexpr operator T() const noexcept {
            return m_value;
        }

        constexpr T operator()() const noexcept {
            return m_value;
        }
    };

    template<bool Value>
    struct BoolConstant : IntegralConstant<bool, Value> {};

    struct FalseType : IntegralConstant<bool, false> {};

    struct TrueType : IntegralConstant<bool, true> {};

    template <class T>
    struct RemoveReference {
        typedef T Type;
    };

    template <class T>
    struct RemoveReference<T&> {
        typedef T Type;
    };

    template <class T>
    struct RemoveReference<T&&> {
        typedef T Type;
    };

    template<class T> struct IsLvalueReference : FalseType {};

    template<class T> struct IsLvalueReference<T&> : TrueType {};

    //

    template <bool Test, class T = void>
    struct EnableIf {};

    template <class T>
    struct EnableIf<true, T> {
        using Type = T;
    };

    template <bool Test, class T = void>
    using EnableIfT = typename EnableIf<Test, T>::Type;

    //

    template <class T, unsigned int N = 0>
    inline constexpr SizeT ExtentV = 0;

    template <class T, SizeT N>
    inline constexpr SizeT ExtentV<T[N], 0> = N;

    template <class T, unsigned int I, SizeT N>
    inline constexpr SizeT ExtentV<T[N], I> = ExtentV<T, I - 1>;

    template <class T, unsigned int I>
    inline constexpr SizeT ExtentV<T[], I> = ExtentV<T, I - 1>;

    template <class T>
    struct RemoveExtent {
        using type = T;
    };

    template <class T, SizeT N>
    struct RemoveExtent<T[N]> {
        using type = T;
    };

    template <class T>
    struct RemoveExtent<T[]> {
        using type = T;
    };

    template <class T>
    using RemoveExtentT = typename RemoveExtent<T>::type;

    template <class T>
    struct RemoveAllExtents {
        using Type = T;
    };

    template <class T, SizeT _Ix>
    struct RemoveAllExtents<T[_Ix]> {
        using Type = typename RemoveAllExtents<T>::Type;
    };

    template <class T>
    struct RemoveAllExtents<T[]> {
        using Type = typename RemoveAllExtents<T>::Type;
    };

    template <class T>
    using RemoveAllExtentsT = typename RemoveAllExtents<T>::Type;

    template <class T>
    struct RemovePointer {
        using Type = T;
    };

    template <class T>
    struct RemovePointer<T*> {
        using Type = T;
    };

    template <class T>
    struct RemovePointer<T* const> {
        using Type = T;
    };

    template <class T>
    struct RemovePointer<T* volatile> {
        using Type = T;
    };

    template <class T>
    struct RemovePointer<T* const volatile> {
        using Type = T;
    };

    template <class T>
    using RemovePointerT = typename RemovePointer<T>::Type;

    // 

    template <class>
    inline constexpr bool IsArrayV = false;

    template <class T, SizeT N>
    inline constexpr bool IsArrayV<T[N]> = true;

    template <class T>
    inline constexpr bool IsArrayV<T[]> = true;

    template <class T>
    struct IsArray : BoolConstant<IsArrayV<T>> {};
}

#endif
