#ifndef __STL__UTILITY_HPP__
#define __STL__UTILITY_HPP__

#include "stl/type_traits.hpp"

#include <utility> // TODO: Remove

namespace stl
{
  template <typename T>
    typename RemoveReference<T>::Type&& move(T&& arg) {
        return static_cast<typename RemoveReference<T>::Type&&>(arg);
    }

    template<typename T> void swap(T& a, T& b) {
        T temp = move(a);
        a = move(b);
        b = move(temp);
    }

    template <class T>
    inline T&& forward(typename RemoveReference<T>::Type& t) noexcept {
        return static_cast<T&&>(t);
    }

    template <class T>
    inline T&& forward(typename RemoveReference<T>::Type&& t) noexcept {
        static_assert(!IsLvalueReference<T>::value
            , "Cannot forward an rvalue as an lvalue.");
        return static_cast<T&&>(t);
    }

    template<typename T, typename TOther> T exchange(T& a, TOther&& b) {
        T temp = move(a);
        a = forward<TOther>(b);
        return move(temp);
    }

  template<typename LeftT, typename RightT>
  using Pair = std::pair<LeftT, RightT>;

#ifdef HRIC_IMPL // TODO: Implement Map
  template<typename LeftT, typename RightT>
  class Pair
  {
  public:
    Pair(const LeftT& left, const RightT& right)
      : m_left(left)
      , m_right(right)
    {}

    Pair(LeftT&& left, const RightT& right)
      : m_left(move(left))
      , m_right(right)
    {}

    Pair(const LeftT& left, RightT&& right)
      : m_left(left)
      , m_right(move(right))
    {}

    Pair(LeftT&& left, RightT&& right)
      : m_left(move(left))
      , m_right(move(right))
    {}

    const LeftT& left() const
    {
      return m_left;
    }

    LeftT& left()
    {
      return m_left;
    }

    const RightT& right() const
    {
      return m_right;
    }

    RightT& right()
    {
      return m_right;
    }

    LeftT m_left;
    RightT m_right;
  };
#endif
}

#endif // __STL__UTILITY_HPP__
