#ifndef __STL_VERSION_HPP__
#define __STL_VERSION_HPP__

#include "stl/stdint.hpp"
#include "stl/string.hpp"

namespace stl {
class Version
{
public:
  using MajorT = stl::UInt8T;
  using MinorT = stl::UInt8T;
  using RevT = stl::UInt8T;

  Version();

  Version(const stl::String& version);

  Version(const MajorT major, const MinorT minor, const RevT revision);

  Version(const Version& other);

  Version(Version&& other) noexcept;

  Version& operator=(const Version& other);

  Version& operator=(Version&& other) noexcept;

  MajorT major() const;

  MinorT minor() const;

  RevT revision() const;

  MajorT& major();

  MinorT& minor();

  RevT& revision();

  bool parse(const stl::String& string);

  String toString() const;

  bool operator<(const Version& other) const;

  bool operator<=(const Version& other) const;

  bool operator>(const Version& other) const;

  bool operator>=(const Version& other) const;

  bool operator==(const Version& other) const;

  bool operator!=(const Version& other) const;

private:
  MajorT m_major;
  MinorT m_minor;
  RevT m_revision;
};
}

#endif // _STL_VERSION_HPP_
