#include "stl/base.hpp"

#include "stl/stl_config.hpp"

#include "log/logging.hpp"

#if (defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1))
#include "stl/network.hpp"
#endif

#if (defined STL_ENABLE_GPIO && (STL_ENABLE_GPIO == 1))
#include "drivers/board/gpio.hpp"
#endif

namespace stl {
InitRetvE
init()
{
#if (defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1))
  if (network::IOStatusE::IOStatusE_OK != network::init()) {
    logging::log << logging::LvlE::Error << "Could not init network library"
                 << stl::endl;
    return InitRetvE::NetworkInitFailed;
  }
#endif

#if (defined STL_ENABLE_GPIO && (STL_ENABLE_GPIO == 1))
  if (!driver::board::GPIO_init()) {
    logging::log << logging::LvlE::Error << "Could not init GPIO library"
                 << stl::endl;
    return InitRetvE::GPIOInitFailed;
  }
#endif

  return InitRetvE::Success;
}
}
