#include "stl/stdint.hpp"
#include "stl/stddef.hpp"

#define BASE64_DEC_LUT_OFFSET 0x2B

static const char g_base64_enc_lut[]
= "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static const char g_base64_dec_lut[] = {
    0x3EU /*+   */,
    0x00U /*,   */, 0x00U /*-   */, 0x00U /*.   */, 0x3FU /*/   */,
    0x34U /*0   */, 0x35U /*1   */, 0x36U /*2   */, 0x37U /*3   */,
    0x38U /*4   */, 0x39U /*5   */, 0x3AU /*6   */, 0x3BU /*7   */,
    0x3CU /*8   */, 0x3DU /*9   */, 0x00U /*:   */, 0x00U /*;   */,
    0x00U /*<   */, 0x00U /*=   */, 0x00U /*>   */, 0x00U /*?   */,
    0x00U /*@   */, 0x00U /*A   */, 0x01U /*B   */, 0x02U /*C   */,
    0x03U /*D   */, 0x04U /*E   */, 0x05U /*F   */, 0x06U /*G   */,
    0x07U /*H   */, 0x08U /*I   */, 0x09U /*J   */, 0x0AU /*K   */,
    0x0BU /*L   */, 0x0CU /*M   */, 0x0DU /*N   */, 0x0EU /*O   */,
    0x0FU /*P   */, 0x10U /*Q   */, 0x11U /*R   */, 0x12U /*S   */,
    0x13U /*T   */, 0x14U /*U   */, 0x15U /*V   */, 0x16U /*W   */,
    0x17U /*X   */, 0x18U /*Y   */, 0x19U /*Z   */, 0x00U /*[   */,
    0x00U /*\\  */, 0x00U /*]   */, 0x00U /*^   */, 0x00U /*_   */,
    0x00U /*`   */, 0x1AU /*a   */, 0x1BU /*b   */, 0x1CU /*c   */,
    0x1DU /*d   */, 0x1EU /*e   */, 0x1FU /*f   */, 0x20U /*g   */,
    0x21U /*h   */, 0x22U /*i   */, 0x23U /*j   */, 0x24U /*k   */,
    0x25U /*l   */, 0x26U /*m   */, 0x27U /*n   */, 0x28U /*o   */,
    0x29U /*p   */, 0x2AU /*q   */, 0x2BU /*r   */, 0x2CU /*s   */,
    0x2DU /*t   */, 0x2EU /*u   */, 0x2FU /*v   */, 0x30U /*w   */,
    0x31U /*x   */, 0x32U /*y   */, 0x33U /*z   */, 0x00U /*{   */,
    0x00U /*|   */, 0x00U /*}   */, 0x00U /*~   */, 0x00U /*DEL */
};

int base64_encode(const stl::UInt8T *input, stl::SizeT length, char *output
    , const stl::SizeT output_buffer_size) {
    const stl::SizeT final_length = ((length / 3U) * 4U
        + ((length % 3U) > 0U ? 4U : 0U));
    if((final_length + 1U) > output_buffer_size) {
        return -1;
    }

    while(length >= 3U) {
        *(output++) = g_base64_enc_lut[0x3F & (*(input) >> 2)];
        *(output++) = g_base64_enc_lut[0x3F & ((*(input) << 4) | (*(input + 1)) >> 4)];
        *(output++) = g_base64_enc_lut[0x3F & ((*(input + 1) << 2) | (*(input + 2)) >> 6)];
        *(output++) = g_base64_enc_lut[0x3F & *(input + 2)];
        input += 3;
        length -= 3U;
    }

    if(1U == length) {
        *(output++) = g_base64_enc_lut[0x3F & (*(input) >> 2)];
        *(output++) = g_base64_enc_lut[0x3F & (*(input) << 4)];
        *(output++) = '=';
        *(output++) = '=';
    } else if(2U == length) {
        *(output++) = g_base64_enc_lut[0x3F & (*(input) >> 2)];
        *(output++) = g_base64_enc_lut[0x3F & ((*(input) << 4) | (*(input + 1)) >> 4)];
        *(output++) = g_base64_enc_lut[0x3F & (*(input + 1) << 2)];
        *(output++) = '=';
    }

    *(output++) = '\0';

    return (int)final_length;
}

/* int base64_decode(const char *input, stl::SizeT length, stl::UInt8T *output
    , const stl::SizeT output_buffer_size) {
    if(0U != (length % 4U)) {
        return -1;
    }

    stl::SizeT final_length = 0U;
    if(length > 0U) {
        final_length = (length / 4U) * 3U;

        if('=' == *(input + length - 1)) {
            --final_length;
        }

        if('=' == *(input + length - 2)) {
            --final_length;
        }

        if(final_length > output_buffer_size) {
            return -1;
        }
    }

    const char *const input_end = input + length;

    while(input_end != input) {
        if('=' == *(input + 1)) {
            *(output++) = (g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *input] << 2);
            break;
        } else {
            *(output++) = (g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *input] << 2)
                | (g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *(input + 1)] >> 4);
        }

        if('=' == *(input + 2)) {
            *(output++) = (g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *(input + 1)] << 4);
            break;
        } else {
            *(output++) = (g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *(input + 1)] << 4)
                | (g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *(input + 2)] >> 2);
        }

        if('=' == *(input + 3)) {
            *(output++) = (g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *(input + 2)] << 6);
            break;
        } else {
            *(output++) = (g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *(input + 2)] << 6)
                | g_base64_dec_lut[BASE64_DEC_LUT_OFFSET + *(input + 3)];
        }

        input += 4;
    }
    return (int)final_length;
} */


int base64_decode(const char *input, stl::SizeT length, stl::UInt8T *output
    , const stl::SizeT output_buffer_size) {
    if(0U != (length % 4U)) {
        return -1;
    }

    stl::SizeT final_length = 0U;
    if(length > 0U) {
        final_length = (length / 4U) * 3U;

        if('=' == *(input + length - 1)) {
            --final_length;
        }

        if('=' == *(input + length - 2)) {
            --final_length;
        }

        if(final_length > output_buffer_size) {
            return -1;
        }
    }

    const char *const input_end = input + length;
    while(input_end != input) {
        if('=' == *(input + 1)) {
            *(output++) = (g_base64_dec_lut[*input - BASE64_DEC_LUT_OFFSET]);
            break;
        } else {
            *(output++) = (g_base64_dec_lut[*input - BASE64_DEC_LUT_OFFSET] << 2)
                | (g_base64_dec_lut[*(input + 1) - BASE64_DEC_LUT_OFFSET] >> 4);
        }

        if('=' == *(input + 2)) {
            *(output++) = (g_base64_dec_lut[*(input + 1) - BASE64_DEC_LUT_OFFSET] << 4);
            break;
        } else {
            *(output++) = (g_base64_dec_lut[*(input + 1) - BASE64_DEC_LUT_OFFSET] << 4)
                | (g_base64_dec_lut[*(input + 2) - BASE64_DEC_LUT_OFFSET] >> 2);
        }

        if('=' == *(input + 3)) {
            *(output++) = (g_base64_dec_lut[*(input + 2) - BASE64_DEC_LUT_OFFSET] << 6);
            break;
        } else {
            *(output++) = (g_base64_dec_lut[*(input + 2) - BASE64_DEC_LUT_OFFSET] << 6)
                | g_base64_dec_lut[*(input + 3) - BASE64_DEC_LUT_OFFSET];
        }

        input += 4;
    }
    return (int)final_length;
}
