#include "crypto/crc.h"

#include "stl/assert.hpp"
#include "stl/stdint.hpp"
#include "stl/stddef.hpp"

/* CRC-32C (iSCSI) polynomial in reversed bit order. */
#define CRC_POLY 0x82F63B78

/* CRC-32 (Ethernet, ZIP, etc.) polynomial in reversed bit order. */
/* #define CRC_POLY 0xEDB88320 */

stl::UInt32T crc32c(stl::UInt32T crc, const stl::UInt8T *data, stl::SizeT size)
{
    stl::Int32T k;

    crc = ~crc;

    while (size--) {
        crc ^= *data++;
        for (k = 0; k < 8; k++) {
            crc = crc & 1 ? (crc >> 1) ^ CRC_POLY : (crc >> 1);
        }
    }

    return ~crc;
}
