#include "crypto/crypto_types.h"

#include "stl/stdint.hpp"
#include "stl/stddef.hpp"

int RC4_ksa(const stl::UInt8T *const key, const stl::UInt32T key_length
    , stl::UInt8T ks[256]) {
    if(NULL == key) {
        return -1;
    }

    if(NULL == ks) {
        return -1;
    }

    for(stl::UInt32T i = 0U; i < 256U; ++i) {
        ks[i] = (stl::UInt8T)i;
    }

    for(stl::UInt32T i = 0U, j = 0U; i < 256U; ++i) {
        j = (j + ks[i] + key[i % key_length]) % 256U;

        const stl::UInt8T t = ks[i];
        ks[i] = ks[j];
        ks[j] = t;
    }
    return 0;
}

int RC4_prga(stl::UInt8T ks[256], const stl::UInt8T *input, stl::SizeT input_size
    , stl::UInt8T *output) {
    if(NULL == ks) {
        return -1;
    }

    if(NULL == input) {
        return -1;
    }

    if(NULL == output) {
        return -1;
    }

    stl::UInt32T i = 0U, j = 0U;

    while(input_size--) {
        i = (i + 1U) % 256U;
        j = (j + ks[i]) % 256U;

        const stl::UInt8T t = ks[i];
        ks[i] = ks[j];
        ks[j] = t;

        *(output++) = *(input++) ^ ks[(ks[i] + ks[j]) % 256U];
    }
    return 0;
}
