#include "drivers/energymeter/eastron_sdm320c.hpp"

#include "log/logging.hpp"

#include "stl/assert.hpp"
#include "stl/compiler.hpp"
#include "stl/error.hpp"
#include "network/packet_builder.hpp"

namespace stl::driver::energymeter {
const EnergymeterDeviceBase::ConnectionTypeListT EastronSDM320CBase::m_connectionTypeList = 
{
  ConnectionTypeE::RS485,
  ConnectionTypeE::EVSE
};

const EnergymeterDeviceBase::UsageListT EastronSDM320CBase::m_usageList =
{
  UsageE::ElectricalNetworkMonitor,
  UsageE::EVSEMonitor
};

EastronSDM320C_RS485::EastronSDM320C_RS485(board::RS485& bus,
                                           BusID busId,
                                           PollingTimeT pollingTime,
                                           const UsageE usage)
  : EastronSDM320CBase(usage, ConnectionTypeE::RS485)
  , EnergymeterRS485Base(bus, stl::move(pollingTime))
  , m_busId(busId)
{
}

stl::ErrorE
EastronSDM320C_RS485::configureHardware()
{
  return stl::ErrorE::Ok;
}

stl::ErrorE
EastronSDM320C_RS485::poll()
{
  try {
    // TODO: Update connectionStatus
  } catch (const std::exception& e) {
    logging::log << logging::LvlE::Exception << '[' << __FILENAME__ << "] "
                 << e.what() << stl::endl;
    return stl::ErrorE::FormatError;
  } catch (...) {
    logging::log << logging::LvlE::Exception << '[' << __FILENAME__
                 << "] Unknown exception caught" << stl::endl;
    return stl::ErrorE::FormatError;
  }

  return stl::ErrorE::Ok;
}
}
