#include "drivers/energymeter/energymeter_device_base.hpp"

#include "log/logging.hpp"

#include "stl/string.hpp"

namespace stl::driver::energymeter {
const stl::String emConnTypeLUT[] = {
#define ENTRY(en, st) st,
  DEF_EM_CONNECTIONTYPE
#undef ENTRY
};

const stl::String emManuLUT[] = {
#define ENTRY(en, st) st,
  DEF_EM_MANUFACTURER
#undef ENTRY
};

const stl::String emModelLUT[] = {
#define ENTRY(en, st) st,
  DEF_EM_MODEL
#undef ENTRY
};

const stl::String emUsageLUT[] = {
#define ENTRY(en, st) st,
  DEF_EM_USAGE
#undef ENTRY
};

const stl::SizeT emUsageLUTItemCount =
  static_cast<stl::SizeT>(stl::driver::energymeter::UsageE::_Count);

const stl::String&
emConnectionTypeToString(stl::driver::energymeter::ConnectionTypeE value)
{
  return emConnTypeLUT[static_cast<stl::SizeT>(value)];
}

stl::driver::energymeter::ConnectionTypeE
stringToEmConnectionType(const stl::String& value)
{
  const stl::SizeT lutItemCount =
    static_cast<stl::SizeT>(stl::driver::energymeter::ConnectionTypeE::_Count);

  for (stl::SizeT i = 0U; i < lutItemCount; ++i) {
    if (0 == emConnTypeLUT[i].ciCompare(value)) {
      return static_cast<stl::driver::energymeter::ConnectionTypeE>(i);
    }
  }

  logging::log << logging::LvlE::Warning << "Invalid connection type \""
               << value << "\"" << stl::endl;

  return stl::driver::energymeter::ConnectionTypeE::Unknown;
}

const stl::String&
emManufacturerToString(stl::driver::energymeter::ManufacturerE value)
{
  return emManuLUT[static_cast<stl::SizeT>(value)];
}

stl::driver::energymeter::ManufacturerE
stringToEmManufacturer(const stl::String& value)
{
  const stl::SizeT lutItemCount =
    static_cast<stl::SizeT>(stl::driver::energymeter::ManufacturerE::_Count);

  for (stl::SizeT i = 0U; i < lutItemCount; ++i) {
    if (0 == emManuLUT[i].ciCompare(value)) {
      return static_cast<stl::driver::energymeter::ManufacturerE>(i);
    }
  }

  logging::log << logging::LvlE::Warning << "Invalid manufacturer \"" << value
               << "\"" << stl::endl;

  return stl::driver::energymeter::ManufacturerE::Unknown;
}

const stl::String&
emModelToString(stl::driver::energymeter::ModelE value)
{
  return emModelLUT[static_cast<stl::SizeT>(value)];
}

stl::driver::energymeter::ModelE
stringToEmModel(const stl::String& value)
{
  const stl::SizeT lutItemCount =
    static_cast<stl::SizeT>(stl::driver::energymeter::ModelE::_Count);

  for (stl::SizeT i = 0U; i < lutItemCount; ++i) {
    if (0 == emModelLUT[i].ciCompare(value)) {
      return static_cast<stl::driver::energymeter::ModelE>(i);
    }
  }

  logging::log << logging::LvlE::Warning << "Invalid model \"" << value << "\""
               << stl::endl;

  return stl::driver::energymeter::ModelE::Unknown;
}

const stl::String&
emUsageToString(stl::driver::energymeter::UsageE value)
{
  Assert(CU32(value) < emUsageLUTItemCount);
  return emUsageLUT[static_cast<stl::SizeT>(value)];
}

stl::driver::energymeter::UsageE
stringToEmUsage(const stl::String& value)
{
  for (stl::SizeT i = 0U; i < emUsageLUTItemCount; ++i) {
    if (0 == emUsageLUT[i].ciCompare(value)) {
      return static_cast<stl::driver::energymeter::UsageE>(i);
    }
  }

  logging::log << logging::LvlE::Warning << "Invalid usage \"" << value << "\""
               << stl::endl;

  return stl::driver::energymeter::UsageE::Unknown;
}
}
