#include "drivers/rfid/rfid_base.hpp"

#include "log/logging.hpp"

#include "stl/string.hpp"

namespace stl::driver::rfid {
const stl::String rfidConnTypeLUT[] = {
#define ENTRY(en, st) st,
  DEF_RFID_CONNECTIONTYPE
#undef ENTRY
};

const stl::String rfidManuLUT[] = {
#define ENTRY(en, st) st,
  DEF_RFID_MANUFACTURER
#undef ENTRY
};

const stl::String rfidModelLUT[] = {
#define ENTRY(en, st) st,
  DEF_RFID_MODEL
#undef ENTRY
};

const stl::String&
rfidConnectionTypeToString(stl::driver::rfid::ConnectionTypeE value)
{
  return rfidConnTypeLUT[static_cast<stl::SizeT>(value)];
}

stl::driver::rfid::ConnectionTypeE
stringToRFIDConnectionType(const stl::String& value)
{
  const stl::SizeT lutItemCount =
    static_cast<stl::SizeT>(stl::driver::rfid::ConnectionTypeE::_Count);

  for (stl::SizeT i = 0U; i < lutItemCount; ++i) {
    if (0 == rfidConnTypeLUT[i].ciCompare(value)) {
      return static_cast<stl::driver::rfid::ConnectionTypeE>(i);
    }
  }

  logging::log << logging::LvlE::Warning << "Invalid connection type \""
               << value << "\"" << stl::endl;

  return stl::driver::rfid::ConnectionTypeE::Unknown;
}

const stl::String&
rfidManufacturerToString(stl::driver::rfid::ManufacturerE value)
{
  return rfidManuLUT[static_cast<stl::SizeT>(value)];
}

stl::driver::rfid::ManufacturerE
stringToRFIDManufacturer(const stl::String& value)
{
  const stl::SizeT lutItemCount =
    static_cast<stl::SizeT>(stl::driver::rfid::ManufacturerE::_Count);

  for (stl::SizeT i = 0U; i < lutItemCount; ++i) {
    if (0 == rfidManuLUT[i].ciCompare(value)) {
      return static_cast<stl::driver::rfid::ManufacturerE>(i);
    }
  }

  logging::log << logging::LvlE::Warning << "Invalid manufacturer \"" << value
               << "\"" << stl::endl;

  return stl::driver::rfid::ManufacturerE::Unknown;
}

const stl::String&
rfidModelToString(stl::driver::rfid::ModelE value)
{
  return rfidModelLUT[static_cast<stl::SizeT>(value)];
}

stl::driver::rfid::ModelE
stringToRFIDModel(const stl::String& value)
{
  const stl::SizeT lutItemCount =
    static_cast<stl::SizeT>(stl::driver::rfid::ModelE::_Count);

  for (stl::SizeT i = 0U; i < lutItemCount; ++i) {
    if (0 == rfidModelLUT[i].ciCompare(value)) {
      return static_cast<stl::driver::rfid::ModelE>(i);
    }
  }

  logging::log << logging::LvlE::Warning << "Invalid model \"" << value << "\""
               << stl::endl;

  return stl::driver::rfid::ModelE::Unknown;
}
}