#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if (defined STL_ENABLE_LOGGING && (STL_ENABLE_LOGGING == 1))

#include "log/logging.hpp"

#include "stl/assert.hpp"
#include "stl/iostream.hpp"
#include "stl/math.hpp"
#include "stl/string_utils.hpp"

// TODO: Make common basic stream for logger and iostream

namespace logging {
Logger log;

static const char*
levelToString(const LvlE level)
{
  Assert(level < LvlE::_Count);

  static const char* const strLUT[] = { "Info",    "Warning",   "Error",
                                        "Success", "Exception", "Debug" };

  if (static_cast<stl::SizeT>(level) >= (sizeof(strLUT) / sizeof(strLUT[0]))) {
    return "";
  }

  return strLUT[static_cast<stl::SizeT>(level)];
}

static stl::ClrE
levelToColor(const LvlE level)
{
  Assert(level < LvlE::_Count);
  Assert(static_cast<stl::SizeT>(level) >= 0U);

  static const stl::ClrE clrLUT[] = { stl::ClrE::Gray,        stl::ClrE::Yellow,
                                 stl::ClrE::LightRed,    stl::ClrE::LightGreen,
                                 stl::ClrE::LightPurple, stl::ClrE::Blue };

  return clrLUT[static_cast<stl::SizeT>(level)];
}

static stl::ClrE
statusToColor(const StatE status)
{
  Assert(status < StatE::_Count);
  Assert(static_cast<stl::SizeT>(status) >= 0U);

  static const stl::ClrE clrLUT[] = { stl::ClrE::Green, stl::ClrE::Red, stl::ClrE::Yellow };

  return clrLUT[static_cast<stl::SizeT>(status)];
}

static const char*
statusToString(const StatE status)
{
  Assert(status < StatE::_Count);
  Assert(static_cast<stl::SizeT>(status) >= 0U);

  static const char* stringLUT[] = { "OK", "Fail", "Warning" };

  return stringLUT[static_cast<stl::SizeT>(status)];
}

stl::SizeT
Logger::write(const stl::UInt8T* const data, stl::SizeT size)
{
  return stl::cout.write(data, size);
}
}

namespace stl::stream {
template<>
OStreamBase&
OStreamBase::operator<<(const logging::LvlE& level)
{
  *this << stl::ClrE::White << '[' << logging::levelToColor(level)
        << logging::levelToString(level) << stl::ClrE::White << "] ";

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const logging::StatE& status)
{
  logging::log << stl::ClrE::White << '[' << logging::statusToColor(status)
               << logging::statusToString(status) << stl::ClrE::White
               << "] ";

  return *this;
}
}

#endif // (defined STL_ENABLE_LOGGING && (STL_ENABLE_LOGGING == 1))
