#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if (defined STL_ENABLE_LOGGING && (STL_ENABLE_LOGGING == 1)) &&               \
  (defined CI_OS_LINUX && (CI_OS_LINUX == 1))

#include "log/logging.hpp"

#include "stl/assert.hpp"
#include "stl/iostream.hpp"

#include <string>

// TODO: Make common basic stream for logger and iostream

static int
colorToCode(const stl::ClrE color)
{
  switch (color) {
    case stl::ClrE::Black:
      return 30;
    case stl::ClrE::Gray:
      return 37;
    case stl::ClrE::Blue:
      return 34;
    case stl::ClrE::LightBlue:
      return 34;
    case stl::ClrE::Green:
      return 32;
    case stl::ClrE::LightGreen:
      return 32;
    case stl::ClrE::Aqua:
      return 36;
    case stl::ClrE::LightAqua:
      return 36;
    case stl::ClrE::Red:
      return 31;
    case stl::ClrE::LightRed:
      return 31;
    case stl::ClrE::Purple:
      return 35;
    case stl::ClrE::LightPurple:
      return 35;
    case stl::ClrE::Yellow:
      return 33;
    case stl::ClrE::LightYellow:
      return 33;
    case stl::ClrE::White:
      return 37;
    case stl::ClrE::BrightWhite:
      return 37;
    case stl::ClrE::_Count:
      break;
    default:
      break;
  }

  return 0;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::ClrE& color)
{
  if ((stl::ClrE::_Count != color) && OutputE::Terminal == m_output) {
    if (m_fgColor != stl::ClrE::_Count) {
      *this << "\x1B[0m";
    }

    *this << "\x1B[" << colorToCode(color) << "m";
  }

  fgColor(color);

  return *this;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream,
           const stl::ClrE& color)
{
  if ((stl::ClrE::_Count != color) &&
      stl::OutputE::Terminal == stream.output()) {
    if (stream.fgColor() != stl::ClrE::_Count) {
      stream << "\x1B[0m";
    }

    stream << "\x1B[" << colorToCode(color) << "m";
  }

  stream.fgColor(color);

  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const std::string& string)
{
  stream.write(string.c_str(), string.length());

  return stream;
}

#endif // #if defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1)
