#include "stl/endian.hpp"

#include "stl/stdint.hpp"

namespace stl {
const EndianE Endian::m_host =
  reinterpret_cast<const UInt8T*>(&Endian::m_test)[0] == 0xCCU
    ? EndianE::EndianE_Big
    : EndianE::EndianE_Little;

template<>
UInt16T
Endian::swap<UInt16T>(const UInt16T value)
{
  return ((value << 8) & 0xFF00U) | ((value >> 8) & 0x00FFU);
}

template<>
Int16T
Endian::swap<Int16T>(const Int16T value)
{
  return ((value << 8) & 0xFF00) | ((value >> 8) & 0x00FF);
}

template<>
UInt32T
Endian::swap<UInt32T>(const UInt32T value)
{
  return ((value << 24) & 0xFF000000U) | ((value >> 24) & 0x000000FFU) |
         ((value << 8) & 0x00FF0000U) | ((value >> 8) & 0x0000FF00U);
}

template<>
Int32T
Endian::swap<Int32T>(const Int32T value)
{
  return ((value << 24) & 0xFF000000) | ((value >> 24) & 0x000000FF) |
         ((value << 8) & 0x00FF0000) | ((value >> 8) & 0x0000FF00);
}

#if defined STL_ENABLE_INT64
template<>
UInt64T
Endian::swap<UInt64T>(const UInt64T value)
{
  return ((value << 56) & 0xFF00000000000000ULL) |
         ((value >> 56) & 0x00000000000000FFULL) |
         ((value << 40) & 0x00FF000000000000ULL) |
         ((value >> 40) & 0x000000000000FF00ULL) |
         ((value << 24) & 0x0000FF0000000000ULL) |
         ((value >> 24) & 0x0000000000FF0000ULL) |
         ((value << 8) & 0x000000FF00000000ULL) |
         ((value >> 8) & 0x00000000FF000000ULL);
}

template<>
Int64T
Endian::swap<Int64T>(const Int64T value)
{
  return ((value << 56) & 0xFF00000000000000LL) |
         ((value >> 56) & 0x00000000000000FFLL) |
         ((value << 40) & 0x00FF000000000000LL) |
         ((value >> 40) & 0x000000000000FF00LL) |
         ((value << 24) & 0x0000FF0000000000LL) |
         ((value >> 24) & 0x0000000000FF0000LL) |
         ((value << 8) & 0x000000FF00000000LL) |
         ((value >> 8) & 0x00000000FF000000LL);
}
#endif
}
