#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#include "stl/fstream.hpp"

#if defined(CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1) && (CI_OS_LINUX == 0)

#include <fstream>

namespace stl::stream {
class FStream::FStreamImpl : FStreamBase
{
public:
  FStreamImpl() = default;

  FStreamImpl(const FStreamImpl&) = delete;
  FStreamImpl& operator=(const FStreamImpl&) = delete;

  FStreamImpl(FStreamImpl&& other) { *this = move(other); }

  FStreamImpl& operator=(FStreamImpl&& other)
  {
    m_stream = ::stl::move(other.m_stream);
    m_error = move(other.m_error);

    return *this;
  }

  FStreamImpl(filesystem::Path path, const FileFlagsT flags)
    : FStreamBase(move(path), flags)
  {
    openInternal(m_path, flags);
  }

  virtual ~FStreamImpl() = default;

  virtual void flush() override { m_stream.flush(); }

  virtual bool open(filesystem::Path path, const FileFlagsT flags) override
  {
    m_path = move(path);
    m_flags = flags;

    return openInternal(
      m_path, flags); // TODO: Reimplement with internal open method in the
                      // parent to move and copy parameters and then pass
                      // them into the internal open method as const refs
  }

  virtual bool isOpen() const override { return m_stream.is_open(); }

  virtual SizeT read(UInt8T* data, SizeT maxSize) override
  {
    return m_stream.read(reinterpret_cast<char*>(data), maxSize).gcount();
  }

  virtual SizeT write(const UInt8T* const data, SizeT size) override
  {
    return m_stream.write(reinterpret_cast<const char*>(data), size).tellp();
  }

  virtual bool seek(const SizeT offset, const SeekModeE seekMode) override
  {
    std::ios_base::seekdir seekdir;

    switch (seekMode) {
      case SeekModeE::Beg:
        seekdir = std::ios::beg;
        break;
      case SeekModeE::Set:
        seekdir = std::ios::beg;
        break;
      case SeekModeE::End:
        seekdir = std::ios::end;
        break;
      default:
        Assert(false);
        return false;
    }

    m_stream.seekg(offset, seekdir);

    return true;
  }

  virtual SizeT position() const override {
    return m_stream.tellp();
  }

  virtual void close() override { m_stream.close(); }

  virtual SizeT size() const override
  {
    const std::streampos position = m_stream.tellp();
    m_stream.seekp(0, std::ios::end);
    const SizeT retv = m_stream.tellp();
    m_stream.seekp(position, std::ios::beg);

    return retv;
  }

  operator bool() const { return isOpen(); }

private:
  bool openInternal(filesystem::Path path, const FileFlagsT flags)
  {
    std::ios::openmode openMode;

    if (flags & FF_R) {
      openMode |= std::ios::in;
    }

    if (flags & FF_W) {
      openMode |= std::ios::out;
    }

    if (flags & FF_App) {
      openMode |= std::ios::app;
    }

    if (flags & FF_Trunc) {
      openMode |= std::ios::trunc;
    }

    m_stream.open(path.c_str(), openMode | std::ios::binary);

    return m_stream.is_open();
  }

  mutable std::fstream m_stream;
};
}
#endif // defined(CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1) && (CI_OS_LINUX == 0)
