#include "stl/stl_config.hpp"
#include "stl/compile_info.hpp"

#if defined CI_PLATFORM_MKRWIFI1010 && (CI_PLATFORM_MKRWIFI1010 == 1)

#include "stl/iostream.hpp"

#include "stl/assert.hpp"
#include "stl/stdint.hpp"
#include "stl/stddef.hpp"
#include "stl/mutex.hpp"

#include "Arduino.h"

stl::OStream& operator << (stl::OStream& stream, const arduino::__FlashStringHelper* const flashString) {
  Assert(flashString);

  stl::LockGuard<stl::Mutex> lock(m_lock);

  Serial.print(flashString);
  Serial.flush();

  return stream;
}

stl::OStream& operator << (stl::OStream& stream, const arduino::StringSumHelper& stringSumHelper) {
  stl::LockGuard<stl::Mutex> lock(m_lock);

  Serial.print(stringSumHelper);
  Serial.flush();

  return stream;
}

namespace stl {
  template<>
  OStream& OStream::operator << (const arduino::__FlashStringHelper* flashString) {
    Assert(flashString);

    stl::LockGuard<stl::Mutex> lock(m_lock);

    Serial.print(flashString);
    Serial.flush();

    return *this;
  }

  template<>
  OStream& OStream::operator << (const arduino::StringSumHelper& stringSumHelper) {
    stl::LockGuard<stl::Mutex> lock(m_lock);

    Serial.print(stringSumHelper);
    Serial.flush();

    return *this;
  }

  template<>
  OStream& OStream::operator << (const EndLineT& value) {
    stl::LockGuard<stl::Mutex> lock(m_lock);

    Serial.println();
    Serial.flush();

    m_floatPrecision = FLOAT_PRECISION_DEFAULT;
    m_integerMode = IntegerModeE::Dec;

    return *this;
  }

  OStream& OStream::write(const char* const str, SizeT length) {
    stl::LockGuard<stl::Mutex> lock(m_lock);

    Serial.write(str, length);
    Serial.flush();

    return *this;
  }

  OStream& OStream::writeAssert(const char* str)
  {
    stl::LockGuard<stl::Mutex> lock(m_lock);

    Serial.write(str, string::length(str));
    Serial.flush();

    return *this;
  }

  OStream& OStream::operator[](const char* message)
  {
    writeAssert(message);

    return *this;
  }
}

#endif
