#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if defined CI_PLATFORM_MKRWIFI1010 && (CI_PLATFORM_MKRWIFI1010 == 1)

#include "stl/memory.hpp"
#include "stl/mutex.hpp"

#include "FreeRTOS.h"
#include "semphr.h"

#include <Arduino.h>

namespace stl {
class MutexImpl
{
public:
  MutexImpl()
    : m_semaphoreHandle(xSemaphoreCreateMutex())
  {
    Assert(m_semaphoreHandle);
  }

  ~MutexImpl()
  {
    Assert(m_semaphoreHandle);
    vSemaphoreDelete(m_semaphoreHandle);
  }

  void lock() { xSemaphoreTake(m_semaphoreHandle, portMAX_DELAY); }

  void unlock() { xSemaphoreGive(m_semaphoreHandle); }

  bool tryLock() { return pdTRUE == xSemaphoreTake(m_semaphoreHandle, 0); }

private:
  xSemaphoreHandle m_semaphoreHandle;
};

Mutex::Mutex()
  : m_handleContainer(makeUnique<MutexImpl>())
{
}

Mutex::~Mutex() {}

void
Mutex::lock()
{
  m_handleContainer->lock();
}

void
Mutex::unlock()
{
  m_handleContainer->unlock();
}

bool
Mutex::tryLock()
{
  return m_handleContainer->tryLock();
}
} // namespace stl

#endif
