#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if (defined CI_SUPPORT_STD && (CI_SUPPORT_STD == 1))

#include "stl/mutex.hpp"

#include <mutex>

namespace stl {
class Mutex::MutexImpl
{
public:
  MutexImpl() = default;

  ~MutexImpl() = default;

  MutexImpl(const MutexImpl&) = delete;
  MutexImpl& operator=(const MutexImpl&) = delete;

  MutexImpl(MutexImpl&&) = delete;
  MutexImpl& operator=(MutexImpl&&) = delete;

  void lock() { m_mutex.lock(); }

  void unlock() { m_mutex.unlock(); }

  bool tryLock() { return m_mutex.try_lock(); }

private:
  std::mutex m_mutex;
};

Mutex::Mutex()
  : m_handleContainer(makeUnique<MutexImpl>())
{
  Assert(m_handleContainer);
}

Mutex::Mutex(String name)
  : m_handleContainer(makeUnique<MutexImpl>())
  , m_name(move(name))
{
  Assert(m_handleContainer);
}

Mutex::~Mutex() {}

void
Mutex::lock()
{
  // Assert(m_handleContainer);
  m_handleContainer->lock();
}

void
Mutex::unlock()
{
  Assert(m_handleContainer);
  m_handleContainer->unlock();
}

bool
Mutex::tryLock()
{
  Assert(m_handleContainer);
  return m_handleContainer->tryLock();
}
} // namespace stl

#endif
