#include "stl/stl_config.hpp"
#include "stl/compile_info.hpp"

#if defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1)

#if defined CI_PLATFORM_MKRWIFI1010 && (CI_PLATFORM_MKRWIFI1010 == 1)

#include "stl/network.hpp"

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0)

#include <ArduinoBearSSL.h>
#include <ArduinoECCX08.h>
#include <WiFi.h>
#include <utility/ECCX08SelfSignedCert.h>

/* static int certVerifyCB(X509_STORE_CTX* ctx, void* arg) { // TODO: Remove
    UNUSED_PARAM(ctx);
    UNUSED_PARAM(arg);
    return 1;
}

static int testErrPrint(const char* str, size_t len, void* u) {
    UNUSED_PARAM(u);

    stl::cout.write(str, len);

    return 1;
};

static int pkPassCB(char* buf, int size, int rwflag, void* userdata) {
    UNUSED_PARAM(rwflag);
    UNUSED_PARAM(userdata);

    static const char pass[] = "D,W0^zjn.i!J[G+c&)Hq";

    if (sizeof(pass) > size) {
        return 0;
    } else {
        strcpy_s(buf, size, pass);

        return sizeof(pass) - 1;
    }WiFi.getTime()
} */

class ArduinoClientSocket : public Client
        {
        public:
            ArduinoClientSocket(::network::ClientSocket& clientSocket)
                : m_clientSocket(&clientSocket) {}

            ArduinoClientSocket(ArduinoClientSocket&& other)
                : m_clientSocket(other.m_clientSocket) {}

            ArduinoClientSocket& operator = (ArduinoClientSocket&& other) {
                m_clientSocket = other.m_clientSocket;

                return *this;
            }

            ArduinoClientSocket(const ArduinoClientSocket&) = delete;
            ArduinoClientSocket& operator = (const ArduinoClientSocket&) = delete;

            virtual int connect(IPAddress ip, uint16_t port) override
            {
            }

            virtual int connect(const char *host, uint16_t port) override
            {
            }

            virtual size_t write(uint8_t) override
            {
            }

            virtual size_t write(const uint8_t *buf, size_t size) override
            {
            }

            virtual int available() override
            {
            }

            virtual int read() override
            {
            }

            virtual int read(uint8_t *buf, size_t size) override
            {
            }

            virtual int peek() override
            {
                return -1;
            }

            virtual void flush() override
            {
            }

            virtual void stop() override
            {
            }

            virtual uint8_t connected() override
            {
            }

            virtual operator bool() override
            {
            }

        private:
            ::network::ClientSocket* m_clientSocket;
        };

namespace network
{
    struct SSLContainer {
        SSLContainer(::network::ClientSocket& clientSocket)
            : m_clientSocket(clientSocket)
            , m_sslClient(m_clientSocket)
            , m_lastSSLError(-1) {}

        SSLContainer(const SSLContainer&) = delete;
        SSLContainer& operator =(const SSLContainer&) = delete;

        SSLContainer(SSLContainer&& other) noexcept {
            operator=(stl::move(other));
        }

        SSLContainer& operator = (SSLContainer&& other) noexcept {
            m_clientSocket = stl::move(other.m_clientSocket);
            m_sslClient = stl::move(m_sslClient);
            m_lastSSLError = stl::exchange(other.m_lastSSLError, -1);

            return *this;
        }

        static SSLContainer* cast(void* const ptr) {
            return reinterpret_cast<SSLContainer*>(ptr);
        }

        static const SSLContainer* cast(const void* const ptr) {
            return reinterpret_cast<const SSLContainer*>(ptr);
        }

        void setLastSSLError(const int error, const char* filename
            , const unsigned int line) {
            m_lastSSLError = m_sslClient.errorCode();

            if (BR_ERR_OK != m_lastSSLError) {
                // TODO: Remove and adjust
                stl::cout << "[SSLError:" << filename << ':' << line << "] "
                    << m_lastSSLError;
            }
        }

        ArduinoClientSocket m_clientSocket;
        BearSSLClient m_sslClient;
        int m_lastSSLError;
    };

    static unsigned long getTime()
    {
        return WiFi.getTime();
    }

    IOStatusE initSecurityLayer()
    {
        ArduinoBearSSL.onGetTime(getTime);

        return IOStatusE::IOStatusE_OK;
    }

    SSLSocketBase::SSLSocketBase()
        : SocketBase()
        , m_sslCtxContainer(nullptr) /* Not needed */ {}

    SSLSocketBase::SSLSocketBase(SocketContainerT &&socketContainer)
        : SocketBase(stl::move(socketContainer)) {}

    SSLSocketBase::SSLSocketBase(SSLSocketBase &&other) noexcept
    {
        operator=(stl::move(other));
    }

    SSLSocketBase &SSLSocketBase::operator=(SSLSocketBase &&other) noexcept
    {
        SocketBase::operator=(stl::move(static_cast<SocketBase &>(other)));
        m_sslCtxContainer = stl::move(other.m_sslCtxContainer);

        return *this;
    }

    int SSLSocketBase::getLastSSLCtxError() const
    {
        return 0;
    }

    void SSLSocketBase::setLastSSLCtxError(const int error, const char *filename, const unsigned int line)
    {}
}

namespace network
{
    SSLClientSocketBase::SSLClientSocketBase()
        : SSLSocketBase()
    {
        m_sslContainer = reinterpret_cast<SSLDataContainer*>(stl::memory::MemoryManager<SSLContainer>::allocate(static_cast<ClientSocket&>(*this)));
            
    }

    SSLClientSocketBase::SSLClientSocketBase(SocketContainerT &&socketContainer, SSLStructContainerT &&sslStructContainer) noexcept
    {
        SSLSocketBase::operator=(stl::move(socketContainer));
        m_sslContainer = stl::move(sslStructContainer);
    }

    SSLClientSocketBase::SSLClientSocketBase(SSLClientSocketBase &&other) noexcept
    {
        SSLClientSocketBase::operator=(stl::move(other));
    }

    SSLClientSocketBase &SSLClientSocketBase::operator=(SSLClientSocketBase &&other) noexcept
    {
        SSLSocketBase::operator=(stl::move(static_cast<SSLSocketBase &>(other)));
        m_sslContainer = stl::move(other.m_sslContainer);

        return *this;
    }

    IOStatusE SSLClientSocketBase::shutdown(const SockDirE dir)
    {
        if (!m_sslContainer)
        {
            static_cast<void>(ClientSocketBase::shutdown(SockDirE::Both));
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (!SSLContainer::cast(m_sslContainer.get())->m_ssl)
        {
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        const int status = SSL_shutdown(SSLContainer::cast(m_sslContainer.get())->m_ssl.get());

        setLastSSLError(status, __FILENAME__, __LINE__);

        return ClientSocketBase::shutdown(dir);
    }

    IOStatusE SSLClientSocketBase::handshake()
    {
        if (!m_sslContainer)
        {
            static_cast<void>(ClientSocketBase::shutdown(SockDirE::Both));
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (!SSLContainer::cast(m_sslContainer.get())->m_ssl)
        {
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        int status = SSL_set_fd(
            SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), static_cast<int>(SocketBase::fd()));

        SSLContainer::cast(m_sslContainer.get())->setLastSSLError(status, __FILENAME__, __LINE__);

        if (1 != status)
        {
            return IOStatusE::IOStatusE_SSLFDError;
        }

        // status = SSL_set_tlsext_host_name(SSLContainer::cast(m_sslContainer.get())->m_ssl, "iont");

        // SSLContainer::cast(m_sslContainer.get())->setLastSSLError(status, __FILENAME__, __LINE__);
        //
        // if (status != 1) {
        //     return IOStatusE::IOStatusE_SSLError;
        // }

        SSL_set_verify(SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), SSL_VERIFY_NONE, nullptr); // TODO: Verify

        status = SSL_accept(SSLContainer::cast(m_sslContainer.get())->m_ssl.get());

        SSLContainer::cast(m_sslContainer.get())
            ->setLastSSLError(status, __FILENAME__, __LINE__);
        setSysError();

        if (status == 1)
        {
            return IOStatusE::IOStatusE_OK;
        }
        else if (0 == status)
        {
            // TODO: SSLContainer::cast(m_sslContainer.get())->m_lastSysError ? For example 10054 connreset
            return IOStatusE::IOStatusE_ConnectionClosed;
        }

        return IOStatusE::IOStatusE_SSLError;
    }

    IOStatusE SSLClientSocketBase::send(
        const stl::UInt8T *const data, const stl::SizeT size, stl::SizeT &sent)
    {
        sent = 0U;

        Assert(nullptr != data);

        if (!m_sslContainer)
        {
            static_cast<void>(ClientSocketBase::shutdown(SockDirE::Both));
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (!SSLContainer::cast(m_sslContainer.get())->m_ssl)
        {
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (size > static_cast<stl::SizeT>(stl::Limits<int>::max()))
        {
            return IOStatusE::IOStatusE_InvalidParam;
        }

        const int retv = SSL_write(
            SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), data, static_cast<int>(size));

        SSLContainer::cast(m_sslContainer.get())
            ->setLastSSLError(retv, __FILENAME__, __LINE__);
        setSysError();

        if (retv > 0)
        {
            sent = static_cast<stl::SizeT>(retv);
            return IOStatusE::IOStatusE_OK;
        }
        else if ((SSL_ERROR_WANT_READ == retv) || (SSL_ERROR_WANT_WRITE == retv))
        {
            return IOStatusE::IOStatusE_OpWouldBlock;
        }

        return IOStatusE::IOStatusE_SSLError;
    }

    IOStatusE SSLClientSocketBase::receive(
        stl::UInt8T *const data, const stl::SizeT maxSize, stl::SizeT &received)
    {
        received = 0U;

        Assert(nullptr != data);

        if (!m_sslContainer)
        {
            static_cast<void>(ClientSocketBase::shutdown(SockDirE::Both));
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (!SSLContainer::cast(m_sslContainer.get())->m_ssl)
        {
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (maxSize > static_cast<stl::SizeT>(stl::Limits<int>::max()))
        {
            return IOStatusE::IOStatusE_InvalidParam;
        }

        const int retv = SSL_read(
            SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), data, static_cast<int>(maxSize));

        SSLContainer::cast(m_sslContainer.get())
            ->setLastSSLError(retv, __FILENAME__, __LINE__);
        setSysError();

        if (retv > 0)
        {
            received = static_cast<stl::SizeT>(retv);
            return IOStatusE::IOStatusE_OK;
        }
        else if ((SSL_ERROR_WANT_READ == retv) || (SSL_ERROR_WANT_WRITE == retv))
        {
            return IOStatusE::IOStatusE_OpWouldBlock;
        }

        return IOStatusE::IOStatusE_SSLError;
    }

    int SSLClientSocketBase::getLastSSLError() const
    {
        if (m_sslContainer)
        {
            return SSLContainer::cast(m_sslContainer.get())->m_lastSSLError;
        }
        SSL_ERROR_NONE;
        return -1;
    }

    void SSLClientSocketBase::setLastSSLError(const int error, const char *filename, const unsigned int line)
    {
        if (m_sslContainer && SSLContainer::cast(m_sslContainer.get())->m_ssl)
        {
            SSLContainer::cast(m_sslContainer.get())->m_lastSSLError = SSL_get_error(
                SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), error);

            if (SSL_ERROR_NONE != SSLContainer::cast(m_sslContainer.get())->m_lastSSLError)
            {
                stl::cout << "[SSLError:" << filename << ':' << line << "] ";
                ERR_print_errors_cb(testErrPrint, nullptr);
            }
        }
        else
        {
            SSLContainer::cast(m_sslContainer.get())->m_lastSSLError = -1;
        }
    }
}

namespace network
{
    SSLClientSocket::SSLClientSocket()
        : SSLClientSocketBase() {}

    IOStatusE SSLClientSocket::postOpenAction() {
        stl::cout << "[OPENSSL]: " << OpenSSL_version(OPENSSL_FULL_VERSION_STRING) << stl::endl;

        if (!m_sslContainer)
        {
            static_cast<void>(SocketBase::shutdown(SockDirE::Both));
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (!m_sslCtxContainer)
        {
            static_cast<void>(SocketBase::shutdown(SockDirE::Both));
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (!SSLContainer::cast(m_sslContainer.get())->m_ssl)
        {
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslMethod = TLS_client_method();

        SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx = SSL_CTX_new(SSLCtxContainer::cast(m_sslCtxContainer.get())
                                                                                   ->m_sslMethod);

        if (!SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx)
        {
            return IOStatusE::IOStatusE_SSLError;
        }

        int status = SSL_CTX_set_ciphersuites(
            SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(), "TLS_AES_256_GCM_SHA384:"
                                                                            "TLS_CHACHA20_POLY1305_SHA256:"
                                                                            "TLS_AES_128_CCM_SHA256:"
                                                                            "TLS_AES_128_CCM_8_SHA256:"
                                                                            "TLS_AES_128_GCM_SHA256"); // ADH-AES128-SHA:ADH-AES128-SHA256:ADH-AES256-SHA256
        if (1 != status)
        {
            return IOStatusE::IOStatusE_SSLCipherError;
        }

        status = SSL_CTX_set_cipher_list(
            SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(), "TLSv1.3:kRSA:!eNULL:!aNULL");
        if (1 != status)
        {
            return IOStatusE::IOStatusE_SSLCipherError;
        }

        // long statusl = SSL_CTX_set_min_proto_version(
        //     sslBaseContextContainer->m_sslCtx, TLS1_2_VERSION);
        // if (1 != statusl) {
        //     return IOStatusE::IOStatusE_SSLContextError;
        // }

        // SSL_CTX_set_options(sslBaseContextContainer->m_sslCtx, SSL_OP_NO_SSLv2
        //     | SSL_OP_NO_SSLv3); // SSL_OP_NO_COMPRESSION

        // SSL_CTX_set_verify(sslBaseContextContainer->m_sslCtx, SSL_VERIFY_NONE
        //     , nullptr);
        // SSL_CTX_set_cert_verify_callback(sslBaseContextContainer->m_sslCtx
        //     , certVerifyCB, NULL);
        // ERR_print_errors_cb(testErrPrint, nullptr);

        // SSL_CTX_set_verify_depth(sslBaseContextContainer->m_sslCtx, 2);

        // int retv = SSL_CTX_load_verify_locations(
        //     sslBaseContextContainer->m_sslCtx, "random-org-chain.pem", NULL);

        // const char PREFERRED_CIPHERS[] = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4";
        // retv = SSL_set_cipher_list(sslBaseContextContainer->m_sslCtx, PREFERRED_CIPHERS);

        // X509* cert = SSL_get_peer_certificate(sslBaseContextContainer->m_sslCtx);
        // X509_free(cert);

        /* Step 2: verify the result of chain verification */
        /* Verification performed according to RFC 4158    */
        // retv = SSL_get_verify_result(ssl);

        // TODO: Add cert support?
        // /* Set the key and cert */
        // if (SSL_CTX_use_certificate_file(sslCtx, "cert.pem", SSL_FILETYPE_PEM) <= 0) {
        //     SSL_CTX_free(sslCtx);
        //     clientSocket.close();
        //     m_stop.store(true);
        //     return;
        // }
        //
        // if (SSL_CTX_use_PrivateKey_file(sslCtx, "key.pem", SSL_FILETYPE_PEM) <= 0) {
        //     SSL_CTX_free(sslCtx);
        //     clientSocket.close();
        //     m_stop.store(true);
        //

        return IOStatusE::IOStatusE_OK;
    }

    SSLClientSocket::SSLClientSocket(SSLClientSocket &&other) noexcept
    {
        operator=(stl::move(other));
    }

    SSLClientSocket &SSLClientSocket::operator=(SSLClientSocket &&other) noexcept
    {
        SSLClientSocketBase::operator=(stl::move(other));

        return *this;
    }

    SSLClientSocket::~SSLClientSocket()
    {
        // TODO
    }

    IOStatusE SSLClientSocket::handshake()
    {
        if (!m_sslContainer)
        {
            static_cast<void>(SocketBase::shutdown(SockDirE::Both));
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        if (!SSLContainer::cast(m_sslContainer.get())->m_ssl)
        {
            return IOStatusE::IOStatusE_SSLNotReady;
        }

        int status = SSL_set_fd(
            SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), static_cast<int>(fd()));

        setLastSSLError(status, __FILENAME__, __LINE__);

        if (1 != status)
        {
            return IOStatusE::IOStatusE_SSLFDError;
        }

        // TODO: Remove
        const char *cipherList = SSL_get_cipher_list(SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), 0);
        stl::cout << "[Client] SSL Version: " << SSL_get_version(SSLContainer::cast(m_sslContainer.get())->m_ssl.get()) << stl::endl;
        stl::cout << "[Client] CipherList: " << cipherList << stl::endl;
        stl::cout << "[CLient] SSL Client Version" << SSL_client_version(SSLContainer::cast(m_sslContainer.get())->m_ssl.get()) << stl::endl;

        status = SSL_connect(
            SSLContainer::cast(m_sslContainer.get())->m_ssl.get());

        setLastSSLError(status, __FILENAME__, __LINE__);
        setSysError();

        if (status == 1)
        {
            return IOStatusE::IOStatusE_OK;
        }
        else if ((SSL_ERROR_WANT_READ == status) || (SSL_ERROR_WANT_WRITE == status))
        {
            return IOStatusE::IOStatusE_OpWouldBlock;
        }
        else
        {
            ERR_print_errors_cb(testErrPrint, nullptr);
        }

        return IOStatusE::IOStatusE_SSLError;
    }

    IOStatusE SSLClientSocket::connectInternal()
    {
        IOStatusE status = ClientSocket::connect(m_networkAddr);

        if (IOStatusE::IOStatusE_OK != status)
        {
            return status;
        }

        SSLContainer::cast(m_sslContainer.get())->m_ssl = SSL_new(SSLCtxContainer::cast(m_sslCtxContainer.get())
                                                                      ->m_sslCtx.get());

        if (!SSLContainer::cast(m_sslContainer.get())->m_ssl)
        {
            SocketBase::shutdown(SockDirE::Both);
            SocketBase::close();
            return IOStatusE::IOStatusE_SSLError;
        }

        // TODO: Remove
        const char *cipherList = SSL_get_cipher_list(SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), 0);
        stl::cout << "[SSL Version]: " << SSL_get_version(SSLContainer::cast(m_sslContainer.get())->m_ssl.get()) << stl::endl;
        stl::cout << "CipherList: " << cipherList << stl::endl;

        return IOStatusE::IOStatusE_OK;
    }

    IOStatusE SSLClientSocket::connect(const NetworkAddress &networkAddress)
    {
        switch (networkAddress.getIPVer())
        {
        case IPVerE::IPv4:
            return connect(networkAddress.getAsIPv4(), networkAddress.getPort());
        case IPVerE::IPv6:
            return connect(networkAddress.getAsIPv6(), networkAddress.getPort());
        default:
            break;
        }

        return IOStatusE::IOStatusE_InvalidParam;
    }

    IOStatusE SSLClientSocket::connect(const IPv4 &ip, const PortT port)
    {
        static_cast<void>(m_networkAddr.set(ip, port)); // It can fail only if parsing string

        return connectInternal();
    }

    IOStatusE SSLClientSocket::connect(const IPv6 &ip, const PortT port)
    {
        static_cast<void>(m_networkAddr.set(ip, port)); // It can fail only if parsing string

        return connectInternal();
    }

    IOStatusE SSLClientSocket::connect(const char *host, const PortT port)
    {
        UNUSED_PARAM(host);
        UNUSED_PARAM(port);

        return IOStatusE::IOStatusE_UnsupportedFunction;
    }
}

#endif // (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0)

#endif // defined CI_PLATFORM_MKRWIFI1010 && (CI_PLATFORM_MKRWIFI1010 == 1)

#endif // defined STL_ENABLE_NETWORKING
