#include "stl/string_utils.hpp"

#include "stl/assert.hpp"
#include "stl/string.hpp"

namespace stl
{
namespace string
{
char* copy(const char* begin, const char* const end, char* destination,
           const stl::SizeT destinationLength)
{
    Assert(nullptr != begin);
    Assert(nullptr != end);
    Assert(nullptr != destination);
    Assert(begin <= end);

    char* const destinationBegin = destination;

    Assert(sizeof(begin) == sizeof(destinationLength));

    if (static_cast<stl::SizeT>(end - begin) >= destinationLength)
    {
        return nullptr;
    }

    while ((begin != end) && (*begin != '\0'))
    {
        *(destination++) = *begin;
    }

    *destination = '\0';

    return destinationBegin;
}

char g_intConvTable[72] =
    "zyxwvutsrqponmlkjihgfedcba9876543210123456789abcdefghijklmnopqrstuvwxyz";

char nibbleToHex(stl::UInt8T nibble, const bool upperCase)
{
    nibble &= 0x0FU;

    if (nibble >= 0x0AU)
    {
        if (upperCase)
        {
            return 'A' + nibble - 0x0AU;
        }
        else
        {
            return 'a' + nibble - 0x0AU;
        }
    }
    else
    {
        return '0' + nibble;
    }
}

void reverse(char* strFront, char* strBack)
{
    Assert(nullptr != strFront);
    Assert(nullptr != strBack);

    while (strFront < strBack)
    {
        const char temp = *strFront;
        *(strFront++) = *strBack;
        *(strBack--) = temp;
    }
}

void reverse(char* str, stl::SizeT length)
{
    Assert(nullptr != str);

    if (length > 0U)
    {
        reverse(str, str + length - 1U);
    }
}

void reverse(char* str)
{
    Assert(nullptr != str);

    reverse(str, length(str));
}

::stl::String toHex(const ::stl::UInt8T* data, const ::stl::SizeT size,
                    bool upperCase)
{
    Assert(nullptr != data);

    ::stl::String result = "";

    for (::stl::SizeT index = 0U; index < size; ++index)
    {
        result += nibbleToHex(*(data + index) >> 4, upperCase);
        result += nibbleToHex(*(data + index), upperCase);
    }

    return result;
}

bool endsWith(const String& input, const char* with)
{
    Assert(with);

    const SizeT withLength = length(with);

    if(input.length() < withLength)
    {
        return false;
    }

    return 0 == input.csCompare(input.length() - withLength, with);
}

bool endsWith(const String& input, const String& with)
{
    if(input.length() < with.length())
    {
        return false;
    }

    return 0 == input.csCompare(input.length() - with.length(), with);
}

bool startsWith(const String& input, const char* with)
{
    Assert(with);

    const SizeT withLength = length(with);

    if(input.length() < withLength)
    {
        return false;
    }

    return 0 == input.ciCompare(0U, with, withLength);
}

bool startsWith(const String& input, const String& with)
{
    if(input.length() < with.length())
    {
        return false;
    }

    return 0 == input.ciCompare(0U, with, with.length());
}

} // namespace string
} // namespace stl
