#include "stl/thread.hpp"

#include "stl/compile_info.hpp"

#if defined(CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1)

#include <chrono>
#include <thread>

namespace stl::thread {
void
yield()
{
  std::this_thread::yield();
}

void
sleepFor(const time::STimeT millis)
{
  std::this_thread::sleep_for(std::chrono::milliseconds(millis));
}

void
sleepForNs(const time::STimeT nanos)
{
  std::this_thread::sleep_for(std::chrono::nanoseconds(nanos));
}

void
sleepForUs(const time::STimeT micros)
{
  std::this_thread::sleep_for(std::chrono::microseconds(micros));
}

void delayForNs(const time::STimeT nanos)
{
  std::chrono::time_point<std::chrono::high_resolution_clock> start = std::chrono::high_resolution_clock::now();
    double elapsed;
    do {
        std::chrono::time_point<std::chrono::high_resolution_clock> end = std::chrono::high_resolution_clock::now();
        elapsed = std::chrono::duration<double, std::nano>(end - start).count();
    } while(elapsed <= nanos);
}
} // namespace stl::thread

#elif defined ARDUINO_SAMD_MKRWIFI1010 && (ARDUINO_SAMD_MKRWIFI1010 == 1)

#include "FreeRTOS.h"
#include "task.h"

#include <Arduino.h>

#include "stl/stdint.hpp"
#include "stl/time.hpp"

namespace stl::thread {
void
yield()
{
  TaskHandle_t handle = xTaskGetCurrentTaskHandle();
  if (nullptr != handle) {
    taskYIELD();
  }
}

void
sleepFor(const time::STimeT millis)
{
  for (time::STimeT ms = 0U; ms < millis; ++ms) {
    yield();
    delay(1);
  }
}
} // namespace stl::thread

#else
#error Unsupported platform
#endif
