#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if defined CI_PLATFORM_MKRWIFI1010 && (CI_PLATFORM_MKRWIFI1010 == 1)

#include "stl/time.hpp"

#include <Arduino.h>

#include "WiFi.h"
#include "NTPClient.h"
#include "utility/wifi_drv.h"

#include "stl/mutex.hpp"

#include "stl/iostream.hpp"

#include "FreeRTOS.h"
#include "task.h"

static WiFiUDP g_wifiUDP;
static NTPClient g_ntpClient(g_wifiUDP);
static stl::Mutex g_lock; // TODO: no global mutex
static stl::time::STimeT g_obtainTime = 0;
static stl::time::STimeT g_initialUTCTimestamp = 0;

namespace stl
{
namespace time
{
Millis getUpTime() { return Millis(xTaskGetTickCount() / portTICK_PERIOD_MS); }

Millis getUTCTimestamp()
{
    STimeT retv = 0;

    stl::LockGuard<stl::Mutex> lock(g_lock);

    if(0 == g_initialUTCTimestamp)
    {
        g_ntpClient.begin();
        if(!g_ntpClient.update())
        {
            return 0;
        }

        g_initialUTCTimestamp = g_ntpClient.getEpochTime();
        g_obtainTime = getUpTime();
        retv = g_initialUTCTimestamp;
    }
    else
    {
        retv = g_initialUTCTimestamp + ((getUpTime() - g_obtainTime) / 1000);
    }

    return retv;
}
} // namespace time
} // namespace stl

#endif // defined CI_PLATFORM_MKRWIFI1010 && (CI_PLATFORM_MKRWIFI1010 == 1)
