#include "stl/stl_config.hpp"
#include "stl/compile_info.hpp"

#if (defined CI_SUPPORT_STD && (CI_SUPPORT_STD == 1))

#include "stl/time.hpp"

#include <chrono>
#include <ctime>
#include <time.h>

#include "log/logging.hpp"

namespace stl {
  namespace time {
    Millis getUpTime() { // TODO: This is not uptime!
        STimeT result = 0;
      #if CI_OS_WINDOWS == 1
        result = static_cast<STimeT>(GetTickCount64());
      #else
        struct timespec ts;

        clock_gettime(CLOCK_MONOTONIC, &ts);

        result = (uint64_t)(ts.tv_nsec / 1000000) + ((uint64_t)ts.tv_sec * 1000ull);
      #endif

      return Millis(result);
    }

    Millis getUTCTimestamp() {
      const auto unix_timestamp = std::chrono::seconds(std::time(NULL));

      return Millis(std::chrono::milliseconds(unix_timestamp).count());
    }

    DateTime::DateTime()
    {
      // Note: Return value of localtime is not threadsafe, because it might be
      // (and will be) reused in subsequent calls to std::localtime!

      time_t t = std::time(NULL);
      struct tm tm = *localtime(&t);

      m_sec = tm.tm_sec;
      m_min = tm.tm_min;
      m_hour = tm.tm_hour;
      m_monthDay = tm.tm_mday;
      m_month = tm.tm_mon + 1;
      m_year = tm.tm_year + 1900;
      m_weekDay = (tm.tm_wday + 6) % 7;
      m_yearDay = tm.tm_yday;
      m_isdst = tm.tm_isdst;
    }
  }
}

#endif // #if (defined CI_OS_LINUX && (CI_OS_LINUX == 1)) || (defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1))
