import Chart from 'chart.js/auto'
import { color } from 'chart.js/helpers';

var datapointsCPM = [
    7,  4,    5,    8,    2,    3,   4,  5, 2, 6,
    7,  9,    3,    5,    6,    1,   8,  2, 9, 2,
    43, 123,  131,  120,  73,   2,   2,  6, 7, 4,
    7,  4,    5,    8,    2,    3,   4,  5, 2, 6,
    82, 119,  125,  118,  121,  117, 32, 6, 7, 4,
    2,  3,    6,    1,    4,    2,   2,  6, 7, 4
    ];

var datapointsSv = [
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    ];

var timer = null;
var myChart = null;

(async function() {
    
const DATA_COUNT = datapointsCPM.length;
const labels = [];

for (let i = 0; i < DATA_COUNT; ++i) {
  labels.push(i.toString());
}

labels.push(DATA_COUNT.toString());

const data = {
  labels: labels,
  datasets: [
    {
      label: 'CPM',
      data: datapointsCPM,
      color: color.red,
      fill: false,
      tension: 0.5,
      yAxisID: 'y',
    },
    {
      label: 'mSv',
      data: datapointsSv,
      color: color.green,
      fill: true,
      tension: 0.5,
      yAxisID: 'y1',
    }
  ]
};

  myChart = new Chart(
    document.getElementById('acquisitions'),
    {
      type: 'line',
      data,
      options: {
          scales: {
              x: {
                display: true,
                min: 0,
                max: 60,
                title: {
                  display: true,
                  text: 'Time [min]'
                }
              },
              y: {
                type: 'linear',
                display: true,
                position: 'left',
                min: 0,
                //max: 500,
                title: {
                  display: true,
                  text: 'CPM [1]'
                }
              },
              y1: {
                type: 'linear',
                display: true,
                position: 'right',
                grid: {
                  drawOnChartArea: false, // only want the grid lines for one axis to show up
                },
                min: 0.0,
                // max: 5.0,
                title: {
                  display: true,
                  text: 'Effective dose [Sv]'
                }
              },
          }
      }
    }
  );
})();

function timerFunction()
{
    if(timer)
    {
        clearInterval(timer);
    }

    timer = setInterval(timerFunction, 1000);

    for(var i = 0; i < datapointsCPM.length; i++)
    {
        datapointsSv[i] = (datapointsCPM[i] * 0.0057);
    }

    var d = new Date(); // for now
    d.getHours(); // => 9
    d.getMinutes(); // =>  30
    d.getSeconds(); // => 51

    document.getElementById('d_time').innerText = d.getHours() + ':' + d.getMinutes() + ':' + d.getSeconds();

    myChart.update();
}

timerFunction();

